#!/usr/bin/python
# -*- coding: utf-8 -*-
#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Sentiment analysis server in Python
"""
from .version import __version__

import logging

logger = logging.getLogger(__name__)

logger.info('Using senpy version: {}'.format(__version__))

from .utils import easy, easy_load, easy_test  # noqa: F401

from .models import *  # noqa: F401,F403
from .plugins import *  # noqa: F401,F403
from .extensions import *  # noqa: F401,F403

__all__ = ['api', 'blueprints', 'cli', 'extensions', 'models', 'plugins']



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from future.utils import iteritems
from .models import Error, Results, Entry, from_string
import logging
logger = logging.getLogger(__name__)

boolean = [True, False]

processors = {
    'string_to_tuple': lambda p: p if isinstance(p, (tuple, list)) else tuple(p.split(','))
}

API_PARAMS = {
    "algorithm": {
        "aliases": ["algorithms", "a", "algo"],
        "required": True,
        "default": 'default',
        "processor": 'string_to_tuple',
        "description": ("Algorithms that will be used to process the request."
                        "It may be a list of comma-separated names."),
    },
    "expanded-jsonld": {
        "@id": "expanded-jsonld",
        "description": "use JSON-LD expansion to get full URIs",
        "aliases": ["expanded", "expanded_jsonld"],
        "options": boolean,
        "required": True,
        "default": False
    },
    "with-parameters": {
        "aliases": ['withparameters',
                    'with_parameters'],
        "description": "include initial parameters in the response",
        "options": boolean,
        "default": False,
        "required": True
    },
    "outformat": {
        "@id": "outformat",
        "aliases": ["o"],
        "default": "json-ld",
        "description": """The data can be semantically formatted (JSON-LD, turtle or n-triples),
given as a list of comma-separated fields (see the fields option) or constructed from a Jinja2
template (see the template option).""",
        "required": True,
        "options": ["json-ld", "turtle", "ntriples"],
    },
    "template": {
        "@id": "template",
        "required": False,
        "description": """Jinja2 template for the result. The input data for the template will
be the results as a dictionary.
For example:

Consider the results before templating:

```
[{
    "@type": "entry",
    "onyx:hasEmotionSet": [],
    "nif:isString": "testing the template",
    "marl:hasOpinion": [
        {
            "@type": "sentiment",
            "marl:hasPolarity": "marl:Positive"
        }
    ]
}]
```


And the template:

```
{% for entry in entries %}
{{ entry["nif:isString"] | upper }},{{entry.sentiments[0]["marl:hasPolarity"].split(":")[1]}}
{% endfor %}
```

The final result would be:

```
TESTING THE TEMPLATE,Positive
```
"""

    },
    "fields": {
        "@id": "fields",
        "required": False,
        "description": """A jmespath selector, that can be used to extract a new dictionary, array or value
from the results.
jmespath is a powerful query language for json and/or dictionaries.
It allows you to change the structure (and data) of your objects through queries.

e.g., the following expression gets a list of `[emotion label, intensity]` for each entry:
`entries[]."onyx:hasEmotionSet"[]."onyx:hasEmotion"[]["onyx:hasEmotionCategory","onyx:hasEmotionIntensity"]`

For more information, see: https://jmespath.org

"""
    },
    "help": {
        "@id": "help",
        "description": "Show additional help to know more about the possible parameters",
        "aliases": ["h"],
        "required": True,
        "options": boolean,
        "default": False
    },
    "verbose": {
        "@id": "verbose",
        "description": "Show all properties in the result",
        "aliases": ["v"],
        "required": True,
        "options": boolean,
        "default": False
    },
    "aliases": {
        "@id": "aliases",
        "description": "Replace JSON properties with their aliases",
        "aliases": [],
        "required": True,
        "options": boolean,
        "default": False
    },
    "emotion-model": {
        "@id": "emotionModel",
        "description": """Emotion model to use in the response.
Senpy will try to convert the output to this model automatically.

Examples: `wna:liking` and `emoml:big6`.
        """,
        "aliases": ["emoModel", "emotionModel"],
        "required": False
    },
    "conversion": {
        "@id": "conversion",
        "description": """How to show the elements that have (not) been converted.

* full: converted and original elements will appear side-by-side
* filtered: only converted elements will be shown
* nested: converted elements will be shown, and they will include a link to the original element
(using `prov:wasGeneratedBy`).
""",
        "required": True,
        "options": ["filtered", "nested", "full"],
        "default": "full"
    }
}

EVAL_PARAMS = {
    "algorithm": {
        "aliases": ["plug", "p", "plugins", "algorithms", 'algo', 'a', 'plugin'],
        "description": "Plugins to evaluate",
        "required": True,
        "help": "See activated plugins in /plugins",
        "processor": API_PARAMS['algorithm']['processor']
    },
    "dataset": {
        "aliases": ["datasets", "data", "d"],
        "description": "Datasets to be evaluated",
        "required": True,
        "help": "See avalaible datasets in /datasets"
    }
}

PLUGINS_PARAMS = {
    "plugin-type": {
        "@id": "pluginType",
        "description": 'What kind of plugins to list',
        "aliases": ["pluginType", "plugin_type"],
        "required": True,
        "default": 'analysisPlugin'
    }
}

WEB_PARAMS = {
    "in-headers": {
        "aliases": ["headers", "inheaders", "inHeaders", "in-headers", "in_headers"],
        "description": "Only include the JSON-LD context in the headers",
        "required": True,
        "default": False,
        "options": boolean
    },
}

CLI_PARAMS = {
    "plugin-folder": {
        "aliases": ["folder", "plugin_folder"],
        "required": True,
        "default": "."
    },
}

NIF_PARAMS = {
    "input": {
        "@id": "input",
        "aliases": ["i"],
        "required": True,
        "help": "Input text"
    },
    "intype": {
        "@id": "intype",
        "description": "input type",
        "aliases": ["t"],
        "required": False,
        "default": "direct",
        "options": ["direct", "url", "file"],
    },
    "informat": {
        "@id": "informat",
        "description": "input format",
        "aliases": ["f"],
        "required": False,
        "default": "text",
        "options": ["text", "json-ld"],
    },
    "language": {
        "@id": "language",
        "description": "language of the input",
        "aliases": ["l"],
        "required": False,
    },
    "prefix": {
        "@id": "prefix",
        "description": "prefix to use for new entities",
        "aliases": ["p"],
        "required": True,
        "default": "",
    },
    "urischeme": {
        "@id": "urischeme",
        "description": "scheme for NIF URIs",
        "aliases": ["u"],
        "required": False,
        "default": "RFC5147String",
        "options": ["RFC5147String", ]
    }
}

BUILTIN_PARAMS = {}

for d in [
        NIF_PARAMS, CLI_PARAMS, WEB_PARAMS, PLUGINS_PARAMS, EVAL_PARAMS,
        API_PARAMS
]:
    for k, v in d.items():
        BUILTIN_PARAMS[k] = v


def parse_params(indict, *specs):
    if not specs:
        specs = [NIF_PARAMS]
    logger.debug("Parsing: {}\n{}".format(indict, specs))
    outdict = indict.copy()
    wrong_params = {}
    for spec in specs:
        for param, options in iteritems(spec):
            for alias in options.get("aliases", []):
                # Replace each alias with the correct name of the parameter
                if alias in indict and alias != param:
                    outdict[param] = indict[alias]
                    del outdict[alias]
                    break
            if param not in outdict:
                if "default" in options:
                    # We assume the default is correct
                    outdict[param] = options["default"]
                elif options.get("required", False):
                    wrong_params[param] = spec[param]
                continue
            if 'processor' in options:
                outdict[param] = processors[options['processor']](outdict[param])
            if "options" in options:
                if options["options"] == boolean:
                    outdict[param] = str(outdict[param]).lower() in ['true', '1', '']
                elif outdict[param] not in options["options"]:
                    wrong_params[param] = spec[param]
    if wrong_params:
        logger.debug("Error parsing: %s", wrong_params)
        message = Error(
            status=400,
            message='Missing or invalid parameters',
            parameters=outdict,
            errors=wrong_params)
        raise message
    return outdict


def get_all_params(plugins, *specs):
    '''Return a list of parameters for a given set of specifications and plugins.'''
    dic = {}
    for s in specs:
        dic.update(s)
    dic.update(get_extra_params(plugins))
    return dic


def get_extra_params(plugins):
    '''Get a list of possible parameters given a list of plugins'''
    params = {}
    extra_params = {}
    for plugin in plugins:
        this_params = plugin.get('extra_params', {})
        for k, v in this_params.items():
            if k not in extra_params:
                extra_params[k] = {}
            extra_params[k][plugin.name] = v
    for k, v in extra_params.items():  # Resolve conflicts
        if len(v) == 1:  # Add the extra options that do not collide
            params[k] = list(v.values())[0]
        else:
            required = False
            aliases = None
            options = None
            default = None
            nodefault = False  # Set when defaults are not compatible

            for plugin, opt in v.items():
                params['{}.{}'.format(plugin, k)] = opt
                required = required or opt.get('required', False)
                newaliases = set(opt.get('aliases', []))
                if aliases is None:
                    aliases = newaliases
                else:
                    aliases = aliases & newaliases
                if 'options' in opt:
                    newoptions = set(opt['options'])
                    options = newoptions if options is None else options & newoptions
                if 'default' in opt:
                    newdefault = opt['default']
                    if newdefault:
                        if default is None and not nodefault:
                            default = newdefault
                        elif newdefault != default:
                            nodefault = True
                            default = None
            # Check for incompatibilities
            if options != set():
                params[k] = {
                    'default': default,
                    'aliases': list(aliases),
                    'required': required,
                    'options': list(options)
                }
    return params


def parse_analyses(params, plugins):
    '''
    Parse the given parameters individually for each plugin, and get a list of the parameters that
    belong to each of the plugins. Each item can then be used in the plugin.analyse_entries method.
    '''
    analysis_list = []
    for i, plugin in enumerate(plugins):
        if not plugin:
            continue
        this_params = filter_params(params, plugin, i)
        parsed = parse_params(this_params, plugin.get('extra_params', {}))
        analysis = plugin.activity(parsed)
        analysis_list.append(analysis)
    return analysis_list


def filter_params(params, plugin, ith=-1):
    '''
    Get the values within params that apply to a plugin.
    More specific names override more general names, in this order:

    <index_order>.parameter > <plugin.name>.parameter > parameter


    Example:

    >>> filter_params({'0.hello': True, 'hello': False}, Plugin(), 0)
    { '0.hello': True, 'hello': True}

    '''
    thisparams = {}
    if ith >= 0:
        ith = '{}.'.format(ith)
    else:
        ith = ""
    for k, v in params.items():
        if ith and k.startswith(str(ith)):
            thisparams[k[len(ith):]] = v
        elif k.startswith(plugin.name):
            thisparams[k[len(plugin.name) + 1:]] = v
        elif k not in thisparams:
            thisparams[k] = v
    return thisparams


def parse_call(params):
    '''
    Return a results object based on the parameters used in a call/request.
    '''
    params = parse_params(params, NIF_PARAMS)
    if params['informat'] == 'text':
        results = Results()
        entry = Entry(nif__isString=params['input'], id='prefix:')  # Use @base
        results.entries.append(entry)
    elif params['informat'] == 'json-ld':
        results = from_string(params['input'], cls=Results)
    else:  # pragma: no cover
        raise NotImplementedError('Informat {} is not implemented'.format(
            params['informat']))
    results.parameters = params
    return results



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import logging
import os

from pkg_resources import parse_version, get_distribution, DistributionNotFound

logger = logging.getLogger(__name__)

MSG = 'GSITK is not (properly) installed.'
IMPORTMSG = '{} Some functions will be unavailable.'.format(MSG)
RUNMSG = '{} Install it to use this function.'.format(MSG)


def raise_exception(*args, **kwargs):
    raise Exception(RUNMSG)


try:
    gsitk_distro = get_distribution("gsitk")
    GSITK_VERSION = parse_version(gsitk_distro.version)

    if not os.environ.get('DATA_PATH'):
        os.environ['DATA_PATH'] = os.environ.get('SENPY_DATA', 'senpy_data')

    from gsitk.datasets.datasets import DatasetManager
    from gsitk.evaluation.evaluation import Evaluation as Eval  # noqa: F401
    from gsitk.evaluation.evaluation import EvalPipeline  # noqa: F401
    from sklearn.pipeline import Pipeline
    modules = locals()
    GSITK_AVAILABLE = True
    datasets = {}
    manager = DatasetManager()

    for item in manager.get_datasets():
        for key in item:
            if key in datasets:
                continue
            properties = item[key]
            properties['@id'] = key
            datasets[key] = properties

    def prepare(ds, *args, **kwargs):
        return manager.prepare_datasets(ds, *args, **kwargs)


except (DistributionNotFound, ImportError) as err:
    logger.debug('Error importing GSITK: {}'.format(err))
    logger.warning(IMPORTMSG)
    GSITK_AVAILABLE = False
    GSITK_VERSION = ()
    DatasetManager = Eval = Pipeline = prepare = raise_exception
    datasets = {}



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import os
import logging

logger = logging.getLogger(__name__)

ROOT = os.path.dirname(__file__)
DEFAULT_FILE = os.path.join(ROOT, 'VERSION')


def read_version(versionfile=DEFAULT_FILE):
    try:
        with open(versionfile) as f:
            return f.read().strip()
    except IOError:  # pragma: no cover
        logger.error('Running an unknown version of senpy. Be careful!.')
        return 'devel'


__version__ = read_version()



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import requests
import json

from senpy.plugins import SentimentBox

ENDPOINT = 'http://www.sentiment140.com/api/bulkClassifyJson'


class Sentiment140(SentimentBox):
    '''Connects to the sentiment140 free API: http://sentiment140.com'''

    author = "@balkian"
    version = '0.2'
    url = "https://github.com/gsi-upm/senpy-plugins-community"
    extra_params = {
        'language': {
            "@id": 'lang_sentiment140',
            'description': 'language of the text',
            'aliases': ['language', 'l'],
            'required': True,
            'default': 'auto',
            'options': ['es', 'en', 'auto']
        }
    }

    classes = ['marl:Positive', 'marl:Neutral', 'marl:Negative']
    binary = True

    def predict_many(self, features, activity):
        lang = activity.params["language"]
        data = []

        for feature in features:
            data.append({'text': feature[0]})

        res = requests.post(ENDPOINT,
                            json.dumps({
                                "language": lang,
                                "data": data
                            }))

        for res in res.json()["data"]:
            polarity = int(res['polarity'])
            neutral_value = 2
            if polarity > neutral_value:
                yield [1, 0, 0]
                continue
            elif polarity < neutral_value:
                yield [0, 0, 1]
                continue
            yield [0, 1, 0]

    test_cases = [
        {
            'entry': {
                'nif:isString': 'I love Titanic'
            },
            'params': {},
            'expected': {
                "nif:isString": "I love Titanic",
                'marl:hasOpinion': [
                    {
                        'marl:hasPolarity': 'marl:Positive',
                    }
                ]
            },
            'responses': [{'url': ENDPOINT,
                           'method': 'POST',
                           'json': {'data': [{'polarity': 4}]}}]
        }
    ]



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from senpy.plugins import EmotionConversionPlugin
from senpy.models import EmotionSet, Emotion, Error

import logging
logger = logging.getLogger(__name__)


class CentroidConversion(EmotionConversionPlugin):
    '''
    This plugin converts emotion annotations from a dimensional model to a
    categorical one, and vice versa. The centroids used in the conversion
    are configurable and appear in the semantic description of the plugin.
    '''
    def __init__(self, info, *args, **kwargs):
        if 'centroids' not in info:
            raise Error('Centroid conversion plugins should provide '
                        'the centroids in their senpy file')
        if 'onyx:doesConversion' not in info:
            if 'centroids_direction' not in info:
                raise Error('Please, provide centroids direction')

            cf, ct = info['centroids_direction']
            info['onyx:doesConversion'] = [{
                'onyx:conversionFrom': cf,
                'onyx:conversionTo': ct
            }, {
                'onyx:conversionFrom': ct,
                'onyx:conversionTo': cf
            }]

        if 'aliases' in info:
            aliases = info['aliases']
            ncentroids = {}
            for k1, v1 in info['centroids'].items():
                nv1 = {}
                for k2, v2 in v1.items():
                    nv1[aliases.get(k2, k2)] = v2
                ncentroids[aliases.get(k1, k1)] = nv1
            info['centroids'] = ncentroids

        super(CentroidConversion, self).__init__(info, *args, **kwargs)

        self.dimensions = set()
        for c in self.centroids.values():
            self.dimensions.update(c.keys())
        self.neutralPoints = self.get("neutralPoints", dict())
        if not self.neutralPoints:
            for i in self.dimensions:
                self.neutralPoints[i] = self.get("neutralValue", 0)

    def _forward_conversion(self, original):
        """Sum the VAD value of all categories found weighted by intensity.
        Intensities are scaled by onyx:maxIntensityValue if it is present, else maxIntensityValue
        is assumed to be one. Emotion entries that do not have onxy:hasEmotionIntensity specified
        are assumed to have maxIntensityValue. Emotion entries that do not have
        onyx:hasEmotionCategory specified are ignored."""
        res = Emotion()
        maxIntensity = float(original.get("onyx:maxIntensityValue", 1))
        for e in original.onyx__hasEmotion:
            category = e.get("onyx:hasEmotionCategory", None)
            if not category:
                continue
            intensity = e.get("onyx:hasEmotionIntensity", maxIntensity) / maxIntensity
            if not intensity:
                continue
            centroid = self.centroids.get(category, None)
            if centroid:
                for dim, value in centroid.items():
                    neutral = self.neutralPoints[dim]
                    if dim not in res:
                        res[dim] = 0
                    res[dim] += (value - neutral) * intensity + neutral
        return res

    def _backwards_conversion(self, original):
        """Find the closest category"""
        centroids = self.centroids
        neutralPoints = self.neutralPoints
        dimensions = self.dimensions

        def distance_k(centroid, original, k):
            # k component of the distance between the value and a given centroid
            return (centroid.get(k, neutralPoints[k]) - original.get(k, neutralPoints[k]))**2

        def distance(centroid):
            return sum(distance_k(centroid, original, k) for k in dimensions)

        distances = {k: distance(centroids[k]) for k in centroids}

        logger.debug('Converting %s', original)
        logger.debug('Centroids: %s', centroids)
        logger.debug('Distances: %s', distances)

        emotion = min(distances, key=lambda x: distances[x])

        result = Emotion(onyx__hasEmotionCategory=emotion)
        result.onyx__algorithmConfidence = distance(centroids[emotion])
        return result

    def convert(self, emotionSet, fromModel, toModel, params):

        cf, ct = self.centroids_direction
        logger.debug(
            '{}\n{}\n{}\n{}'.format(emotionSet, fromModel, toModel, params))
        e = EmotionSet()
        if fromModel == cf and toModel == ct:
            e.onyx__hasEmotion.append(self._forward_conversion(emotionSet))
        elif fromModel == ct and toModel == cf:
            for i in emotionSet.onyx__hasEmotion:
                e.onyx__hasEmotion.append(self._backwards_conversion(i))
        else:
            raise Error('EMOTION MODEL NOT KNOWN. '
                        'Cannot convert from {} to {}'.format(fromModel,
                                                              toModel))
        yield e

    def test(self, info=None):
        if not info:
            info = {
                "name": "CentroidTest",
                "description": "Centroid test",
                "version": 0,
                "centroids": {
                    "c1": {"V1": 0.5,
                           "V2": 0.5},
                    "c2": {"V1": -0.5,
                           "V2": 0.5},
                    "c3": {"V1": -0.5,
                           "V2": -0.5},
                    "c4": {"V1": 0.5,
                           "V2": -0.5}},
                "aliases": {
                    "V1": "X-dimension",
                    "V2": "Y-dimension"
                },
                "centroids_direction": ["emoml:big6", "emoml:fsre-dimensions"]
            }

        c = CentroidConversion(info)

        es1 = EmotionSet()
        e1 = Emotion()
        e1.onyx__hasEmotionCategory = "c1"
        es1.onyx__hasEmotion.append(e1)
        res = c._forward_conversion(es1)
        assert res["X-dimension"] == 0.5
        assert res["Y-dimension"] == 0.5

        e2 = Emotion()
        e2.onyx__hasEmotionCategory = "c2"
        es1.onyx__hasEmotion.append(e2)
        res = c._forward_conversion(es1)
        assert res["X-dimension"] == 0
        assert res["Y-dimension"] == 1

        e = Emotion()
        e["X-dimension"] = -0.2
        e["Y-dimension"] = -0.3
        res = c._backwards_conversion(e)
        assert res["onyx:hasEmotionCategory"] == "c3"

        e = Emotion()
        e["X-dimension"] = -0.2
        e["Y-dimension"] = 0.3
        res = c._backwards_conversion(e)
        assert res["onyx:hasEmotionCategory"] == "c2"



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
'''
Senpy Models.

This implementation should mirror the JSON schema definition.
For compatibility with Py3 and for easier debugging, this new version drops
introspection and adds all arguments to the models.
'''
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

from future.utils import with_metaclass
from past.builtins import basestring

from jinja2 import Environment, BaseLoader

import time
import copy
import json
import os
import jsonref
from flask import Response as FlaskResponse
from pyld import jsonld

import logging
import jmespath

logging.getLogger('rdflib').setLevel(logging.WARN)
logger = logging.getLogger(__name__)

from rdflib import Graph


from .meta import BaseMeta, CustomDict, alias

DEFINITIONS_FILE = 'definitions.json'
CONTEXT_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            'schemas',
                            'context.jsonld')


def get_schema_path(schema_file, absolute=False):
    if absolute:
        return os.path.realpath(schema_file)
    else:
        return os.path.join(
            os.path.dirname(os.path.realpath(__file__)), 'schemas',
            schema_file)


def read_schema(schema_file, absolute=False):
    schema_path = get_schema_path(schema_file, absolute)
    schema_uri = 'file://{}'.format(schema_path)
    with open(schema_path) as f:
        return jsonref.load(f, base_uri=schema_uri)


def dump_schema(schema):
    return jsonref.dumps(schema)


def load_context(context):
    logging.debug('Loading context: {}'.format(context))
    if not context:
        return context
    elif isinstance(context, list):
        contexts = []
        for c in context:
            contexts.append(load_context(c))
        return contexts
    elif isinstance(context, dict):
        return dict(context)
    elif isinstance(context, basestring):
        try:
            with open(context) as f:
                return dict(json.loads(f.read()))
        except IOError:
            return context
    else:
        raise AttributeError('Please, provide a valid context')


base_context = load_context(CONTEXT_PATH)


def register(rsubclass, rtype=None):
    BaseMeta.register(rsubclass, rtype)


class BaseModel(with_metaclass(BaseMeta, CustomDict)):
    '''
    Entities of the base model are a special kind of dictionary that emulates
    a JSON-LD object. The structure of the dictionary is checked via JSON-schema.
    For convenience, the values can also be accessed as attributes
    (a la Javascript). e.g.:

    >>> myobject.key == myobject['key']
    True
    >>> myobject.ns__name == myobject['ns:name']
    True

    Additionally, subclasses of this class can specify default values for their
    instances. These defaults are inherited by subclasses. e.g.:

    >>> class NewModel(BaseModel):
    ...     mydefault = 5
    >>> n1 = NewModel()
    >>> n1['mydefault'] == 5
    True
    >>> n1.mydefault = 3
    >>> n1['mydefault'] = 3
    True
    >>> n2 = NewModel()
    >>> n2 == 5
    True
    >>> class SubModel(NewModel):
            pass
    >>> subn = SubModel()
    >>> subn.mydefault == 5
    True

    Lastly, every subclass that also specifies a schema will get registered, so it
    is possible to deserialize JSON and get the right type.
    i.e. to recover an instance of the original class from a plain JSON.

    '''

    # schema_file = DEFINITIONS_FILE
    _context = base_context["@context"]

    def __init__(self, *args, **kwargs):
        auto_id = kwargs.pop('_auto_id', False)

        super(BaseModel, self).__init__(*args, **kwargs)

        if auto_id:
            self.id

    @property
    def id(self):
        if '@id' not in self:
            self['@id'] = 'prefix:{}_{}'.format(type(self).__name__, time.time())
        return self['@id']

    @id.setter
    def id(self, value):
        self['@id'] = value

    def flask(self,
              in_headers=False,
              headers=None,
              outformat='json-ld',
              **kwargs):
        """
        Return the values and error to be used in flask.
        So far, it returns a fixed context. We should store/generate different
        contexts if the plugin adds more aliases.
        """
        headers = headers or {}
        kwargs["with_context"] = not in_headers
        content, mimetype = self.serialize(format=outformat,
                                           with_mime=True,
                                           **kwargs)

        if outformat == 'json-ld' and in_headers:
            headers.update({
                "Link":
                ('<%s>;'
                    'rel="http://www.w3.org/ns/json-ld#context";'
                    ' type="application/ld+json"' % kwargs.get('context_uri'))
            })
        return FlaskResponse(
            response=content,
            status=self.get('status', 200),
            headers=headers,
            mimetype=mimetype)

    def serialize(self, format='json-ld', with_mime=False,
                  template=None, prefix=None, fields=None, **kwargs):
        js = self.jsonld(prefix=prefix, **kwargs)
        if template is not None:
            rtemplate = Environment(loader=BaseLoader).from_string(template)
            content = rtemplate.render(**self)
            mimetype = 'text'
        elif fields is not None:
            # Emulate field selection by constructing a template
            content = json.dumps(jmespath.search(fields, js))
            mimetype = 'text'
        elif format == 'json-ld':
            content = json.dumps(js, indent=2, sort_keys=True)
            mimetype = "application/json"
        elif format in ['turtle', 'ntriples']:
            content = json.dumps(js, indent=2, sort_keys=True)
            logger.debug(js)
            context = [self._context, {'prefix': prefix, '@base': prefix}]
            g = Graph().parse(
                data=content,
                format='json-ld',
                prefix=prefix,
                context=context)
            logger.debug(
                'Parsing with prefix: {}'.format(kwargs.get('prefix')))
            content = g.serialize(format=format,
                                  prefix=prefix)
            mimetype = 'text/{}'.format(format)
        else:
            raise Error('Unknown outformat: {}'.format(format))
        if with_mime:
            return content, mimetype
        else:
            return content

    def jsonld(self,
               with_context=False,
               context_uri=None,
               prefix=None,
               base=None,
               expanded=False,
               **kwargs):

        result = self.serializable(**kwargs)

        if expanded:
            result = jsonld.expand(
                result,
                options={
                    'expandContext': [
                        self._context,
                        {
                            'prefix': prefix,
                            'endpoint': prefix
                        }
                    ]
                }
            )[0]
        if not with_context:
            try:
                del result['@context']
            except KeyError:
                pass
        elif context_uri:
            result['@context'] = context_uri
        else:
            result['@context'] = self._context

        return result

    def validate(self, obj=None):
        if not obj:
            obj = self
        if hasattr(obj, "jsonld"):
            obj = obj.jsonld()
        self._validator.validate(obj)

    def prov(self, another):
        self['prov:wasGeneratedBy'] = another.id


def subtypes():
    return BaseMeta._subtypes


def from_dict(indict, cls=None, warn=True):
    if not cls:
        target = indict.get('@type', None)
        cls = BaseModel
        try:
            cls = subtypes()[target]
        except KeyError:
            pass

    if cls == BaseModel and warn:
        logger.warning('Created an instance of an unknown model')

    outdict = dict()
    for k, v in indict.items():
        if k == '@context':
            pass
        elif isinstance(v, dict):
            v = from_dict(indict[k])
        elif isinstance(v, list):
            v = v[:]
            for ix, v2 in enumerate(v):
                if isinstance(v2, dict):
                    v[ix] = from_dict(v2)
        outdict[k] = copy.copy(v)
    return cls(**outdict)


def from_string(string, **kwargs):
    return from_dict(json.loads(string), **kwargs)


def from_json(injson, **kwargs):
    indict = json.loads(injson)
    return from_dict(indict, **kwargs)


class Entry(BaseModel):
    schema = 'entry'

    text = alias('nif:isString')
    sentiments = alias('marl:hasOpinion', [])
    emotions = alias('onyx:hasEmotionSet', [])


class Sentiment(BaseModel):
    schema = 'sentiment'

    polarity = alias('marl:hasPolarity')
    polarityValue = alias('marl:polarityValue')


class Error(BaseModel, Exception):
    schema = 'error'

    def __init__(self, message='Generic senpy exception', *args, **kwargs):
        Exception.__init__(self, message)
        super(Error, self).__init__(*args, **kwargs)
        self.message = message

    def __str__(self):
        if not hasattr(self, 'errors'):
            return self.message
        return '{}:\n\t{}'.format(self.message, self.errors)

    def __hash__(self):
        return Exception.__hash__(self)


class AggregatedEvaluation(BaseModel):
    schema = 'aggregatedEvaluation'

    evaluations = alias('senpy:evaluations', [])


class Dataset(BaseModel):
    schema = 'dataset'


class Datasets(BaseModel):
    schema = 'datasets'

    datasets = []


class Emotion(BaseModel):
    schema = 'emotion'


class EmotionConversion(BaseModel):
    schema = 'emotionConversion'


class EmotionConversionPlugin(BaseModel):
    schema = 'emotionConversionPlugin'


class EmotionAnalysis(BaseModel):
    schema = 'emotionAnalysis'


class EmotionModel(BaseModel):
    schema = 'emotionModel'
    onyx__hasEmotionCategory = []


class EmotionPlugin(BaseModel):
    schema = 'emotionPlugin'


class EmotionSet(BaseModel):
    schema = 'emotionSet'

    onyx__hasEmotion = []


class Evaluation(BaseModel):
    schema = 'evaluation'

    metrics = alias('senpy:metrics', [])


class Entity(BaseModel):
    schema = 'entity'


class Help(BaseModel):
    schema = 'help'


class Metric(BaseModel):
    schema = 'metric'


class Parameter(BaseModel):
    schema = 'parameter'


class Plugins(BaseModel):
    schema = 'plugins'

    plugins = []


class Response(BaseModel):
    schema = 'response'


class Results(BaseModel):
    schema = 'results'

    _terse_keys = ['entries', ]

    activities = []
    entries = []

    def activity(self, id):
        for i in self.activities:
            if i.id == id:
                return i
        return None


class SentimentPlugin(BaseModel):
    schema = 'sentimentPlugin'


class Suggestion(BaseModel):
    schema = 'suggestion'


class Topic(BaseModel):
    schema = 'topic'


class Analysis(BaseModel):
    '''
    A prov:Activity that results of executing a Plugin on an entry with a set of
    parameters.
    '''
    schema = 'analysis'

    parameters = alias('prov:used', [])
    algorithm = alias('prov:wasAssociatedWith', [])

    @property
    def params(self):
        outdict = {}
        outdict['algorithm'] = self.algorithm
        for param in self.parameters:
            outdict[param['name']] = param['value']
        return outdict

    @params.setter
    def params(self, value):
        for k, v in value.items():
            for param in self.parameters:
                if param.name == k:
                    param.value = v
                    break
            else:
                self.parameters.append(Parameter(name=k, value=v))  # noqa: F821

    def param(self, key, default=None):
        for param in self.parameters:
            if param['name'] == key:
                return param['value']
        return default

    @property
    def plugin(self):
        return self._plugin

    @plugin.setter
    def plugin(self, value):
        self._plugin = value
        self['prov:wasAssociatedWith'] = value.id

    def run(self, request):
        return self.plugin.process(request, self)


class Plugin(BaseModel):
    schema = 'plugin'
    extra_params = {}

    def activity(self, parameters=None):
        '''Generate an Analysis (prov:Activity) from this plugin and the given parameters'''
        a = Analysis()
        a.plugin = self
        if parameters:
            a.params = parameters
        return a


# More classes could be added programmatically

def _class_from_schema(name, schema=None, schema_file=None, base_classes=None):
    base_classes = base_classes or []
    base_classes.append(BaseModel)
    attrs = {}
    if schema:
        attrs['schema'] = schema
    elif schema_file:
        attrs['schema_file'] = schema_file
    else:
        attrs['schema'] = name
    name = "".join((name[0].upper(), name[1:]))
    return BaseMeta(name, base_classes, attrs)


def _add_class_from_schema(*args, **kwargs):
    generatedClass = _class_from_schema(*args, **kwargs)
    globals()[generatedClass.__name__] = generatedClass
    del generatedClass



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from future import standard_library
standard_library.install_aliases()

from future.utils import with_metaclass
from functools import partial

import os.path
import os
import re
import pickle
import logging
import pprint

import inspect
import sys
import subprocess
import importlib
import yaml
import threading
import multiprocessing
import pkg_resources
from nltk import download
from textwrap import dedent
from sklearn.base import TransformerMixin, BaseEstimator
from itertools import product

from .. import models, utils
from .. import api
from .. import gsitk_compat
from .. import testing

logger = logging.getLogger(__name__)


class PluginMeta(models.BaseMeta):
    _classes = {}

    def __new__(mcs, name, bases, attrs, **kwargs):
        plugin_type = set()
        for base in bases:
            if hasattr(base, '_plugin_type'):
                plugin_type |= base._plugin_type
        plugin_type.add(name)
        alias = attrs.get('name', name).lower()
        attrs['_plugin_type'] = plugin_type
        logger.debug('Adding new plugin class: %s %s %s %s', name, bases, attrs, plugin_type)
        attrs['name'] = alias
        if 'description' not in attrs:
            doc = attrs.get('__doc__', None)
            if doc:
                attrs['description'] = dedent(doc)
            else:
                logger.warning(
                    ('Plugin {} does not have a description. '
                     'Please, add a short summary to help other developers'
                     ).format(name))
        cls = super(PluginMeta, mcs).__new__(mcs, name, bases, attrs)

        if alias in mcs._classes:
            if os.environ.get('SENPY_TESTING', ""):
                raise Exception(
                    ('The type of plugin {} already exists. '
                     'Please, choose a different name').format(name))
            else:
                logger.warning('Overloading plugin class: {}'.format(alias))
        mcs._classes[alias] = cls
        return cls

    @classmethod
    def for_type(cls, ptype):
        return cls._classes[ptype]


class Plugin(with_metaclass(PluginMeta, models.Plugin)):
    '''
    Base class for all plugins in senpy.
    A plugin must provide at least these attributes:

        - version
        - description (or docstring)
        - author

    Additionally, they may provide a URL (url) of a repository or website.

    '''

    _terse_keys = ['name', '@id', '@type', 'author', 'description',
                   'extra_params', 'is_activated', 'url', 'version']

    def __init__(self, info=None, data_folder=None, **kwargs):
        """
        Provides a canonical name for plugins and serves as base for other
        kinds of plugins.
        """
        logger.debug("Initialising %s", info)
        super(Plugin, self).__init__(**kwargs)
        if info:
            self.update(info)
        self.validate()
        self.id = 'endpoint:plugins/{}_{}'.format(self['name'],
                                                  self['version'])
        self.is_activated = False
        self._lock = threading.Lock()
        self._directory = os.path.abspath(
            os.path.dirname(inspect.getfile(self.__class__)))

        data_folder = data_folder or os.getcwd()
        subdir = os.path.join(data_folder, self.name)

        self._data_paths = [
            data_folder,
            subdir,
            self._directory,
            os.path.join(self._directory, 'data'),
        ]

        if os.path.exists(subdir):
            data_folder = subdir
        self.data_folder = data_folder

        self._log = logging.getLogger('{}.{}'.format(__name__, self.name))

    @property
    def log(self):
        return self._log

    def validate(self):
        missing = []
        for x in ['name', 'description', 'version']:
            if x not in self:
                missing.append(x)
        if missing:
            raise models.Error(
                'Missing configuration parameters: {}'.format(missing))

    def get_folder(self):
        return os.path.dirname(inspect.getfile(self.__class__))

    def _activate(self):
        self.activate()
        self.is_activated = True

    def _deactivate(self):
        self.is_activated = False
        self.deactivate()

    def activate(self):
        pass

    def deactivate(self):
        pass

    def process(self, request, activity, **kwargs):
        """
        An implemented plugin should override this method.
        Here, we assume that a process_entries method exists.
        """
        newentries = list(
            self.process_entries(request.entries, activity))
        request.entries = newentries
        return request

    def process_entries(self, entries, activity):
        for entry in entries:
            self.log.debug('Processing entry with plugin %s: %s', self, entry)
            results = self.process_entry(entry, activity)
            if inspect.isgenerator(results):
                for result in results:
                    yield result
            else:
                yield results

    def process_entry(self, entry, activity):
        """
        This base method is here to adapt plugins which only
        implement the *process* function.
        Note that this method may yield an annotated entry or a list of
        entries (e.g. in a tokenizer)
        """
        raise NotImplementedError(
            'You need to implement process, process_entries or process_entry in your plugin'
        )

    def test(self, test_cases=None):
        if not test_cases:
            if not hasattr(self, 'test_cases'):
                raise AttributeError(
                    ('Plugin {} [{}] does not have any defined '
                     'test cases').format(self.id,
                                          inspect.getfile(self.__class__)))
            test_cases = self.test_cases
        for case in test_cases:
            try:
                fmt = 'case: {}'.format(case.get('name', case))
                if 'name' in case:
                    self.log.info('Test case: {}'.format(case['name']))
                self.log.debug('Test case:\n\t{}'.format(
                    pprint.pformat(fmt)))
                self.test_case(case)
            except Exception as ex:
                self.log.warning('Test case failed:\n{}'.format(
                    pprint.pformat(case)))
                raise

    def test_case(self, case, mock=testing.MOCK_REQUESTS):
        if 'entry' not in case and 'input' in case:
            entry = models.Entry(_auto_id=False)
            entry.nif__isString = case['input']
            case['entry'] = entry
        entry = models.Entry(case['entry'])
        given_parameters = case.get('params', case.get('parameters', {}))
        expected = case.get('expected', None)
        should_fail = case.get('should_fail', False)
        responses = case.get('responses', [])

        try:
            request = models.Response()
            parameters = api.parse_params(given_parameters,
                                          self.extra_params)
            request.entries = [
                entry,
            ]

            activity = self.activity(parameters)

            method = partial(self.process, request, activity)

            if mock:
                res = method()
            else:
                with testing.patch_all_requests(responses):
                    res = method()

            if not isinstance(expected, list):
                expected = [expected]
            utils.check_template(res.entries, expected)
            res.validate()
        except models.Error:
            if should_fail:
                return
            raise
        assert not should_fail

    def find_file(self, fname):
        for p in self._data_paths:
            alternative = os.path.join(p, fname)
            if os.path.exists(alternative):
                return alternative
        raise IOError('File does not exist: {}'.format(fname))

    def path(self, fpath):
        if not os.path.isabs(fpath):
            fpath = os.path.join(self.data_folder, fpath)
        return fpath

    def open(self, fpath, mode='r'):
        if 'w' in mode:
            # When writing, only use absolute paths or data_folder
            fpath = self.path(fpath)
        else:
            fpath = self.find_file(fpath)

        return open(fpath, mode=mode)

    def serve(self, debug=True, **kwargs):
        utils.easy(plugin_list=[self, ], plugin_folder=None, debug=debug, **kwargs)


# For backwards compatibility
SenpyPlugin = Plugin


class Analyser(Plugin):
    '''
    A subclass of Plugin that analyses text and provides an annotation.
    '''

    # Deprecated
    def analyse(self, request, activity):
        return super(Analyser, self).process(request, activity)

    # Deprecated
    def analyse_entries(self, entries, activity):
        for i in super(Analyser, self).process_entries(entries, activity):
            yield i

    def process(self, request, activity, **kwargs):
        return self.analyse(request, activity)

    def process_entries(self, entries, activity):
        for i in self.analyse_entries(entries, activity):
            yield i

    def process_entry(self, entry, activity, **kwargs):
        if hasattr(self, 'analyse_entry'):
            for i in self.analyse_entry(entry, activity):
                yield i
        else:
            super(Analyser, self).process_entry(entry, activity, **kwargs)


AnalysisPlugin = Analyser


class Transformation(AnalysisPlugin):
    '''Empty'''
    pass


class Conversion(Plugin):
    '''
    A subclass of Plugins that convert between different annotation models.
    e.g. a conversion of emotion models, or normalization of sentiment values.
    '''

    def process(self, response, parameters, plugins=None, **kwargs):
        plugins = plugins or []
        newentries = []
        for entry in response.entries:
            newentries.append(
                self.convert_entry(entry, parameters, plugins))
        response.entries = newentries
        return response

    def convert_entry(self, entry, parameters, conversions_applied):
        raise NotImplementedError(
            'You should implement a way to convert each entry, or a custom process method'
        )


ConversionPlugin = Conversion


class Evaluable(Plugin):
    '''
    Common class for plugins that can be evaluated with GSITK.

    They should implement the methods below.
    '''

    def as_pipe(self):
        raise Exception('Implement the as_pipe function')

    def evaluate_func(self, X, activity=None):
        raise Exception('Implement the evaluate_func function')

    def evaluate(self, *args, **kwargs):
        return evaluate([self], *args, **kwargs)


class SentimentPlugin(Analyser, Evaluable, models.SentimentPlugin):
    '''
    Sentiment plugins provide sentiment annotation (using Marl)
    '''
    minPolarityValue = 0
    maxPolarityValue = 1

    _terse_keys = Analyser._terse_keys + ['minPolarityValue', 'maxPolarityValue']

    def test_case(self, case):
        if 'polarity' in case:
            expected = case.get('expected', {})
            s = models.Sentiment(_auto_id=False)
            s.marl__hasPolarity = case['polarity']
            if 'marl:hasOpinion' not in expected:
                expected['marl:hasOpinion'] = []
            expected['marl:hasOpinion'].append(s)
            case['expected'] = expected
        super(SentimentPlugin, self).test_case(case)

    def normalize(self, value, minValue, maxValue):
        nv = minValue + (value - self.minPolarityValue) * (
            self.maxPolarityValue - self.minPolarityValue) / (maxValue - minValue)
        return nv

    def as_pipe(self):
        pipe = gsitk_compat.Pipeline([('senpy-plugin', ScikitWrapper(self))])
        pipe.name = self.id
        return pipe

    def evaluate_func(self, X, activity=None):
        if activity is None:
            parameters = api.parse_params({},
                                          self.extra_params)
            activity = self.activity(parameters)
        entries = []
        for feat in X:
            if isinstance(feat, list):
                feat = ' '.join(feat)
            entries.append(models.Entry(nif__isString=feat))
        labels = []
        for e in self.process_entries(entries, activity):
            sent = e.sentiments[0].polarity
            label = -1
            if sent == 'marl:Positive':
                label = 1
            elif sent == 'marl:Negative':
                label = -1
            labels.append(label)
        return labels


class EmotionPlugin(Analyser, models.EmotionPlugin):
    '''
    Emotion plugins provide emotion annotation (using Onyx)
    '''
    minEmotionValue = 0
    maxEmotionValue = 1

    _terse_keys = Analyser._terse_keys + ['minEmotionValue', 'maxEmotionValue']


class EmotionConversion(Conversion):
    '''
    A subclass of Conversion that converts emotion annotations using different models
    '''

    def can_convert(self, fromModel, toModel):
        '''
        Whether this plugin can convert from fromModel to toModel.
        If fromModel is None, it is interpreted as "any Model"
        '''
        for pair in self.onyx__doesConversion:
            if (pair['onyx:conversionTo'] == toModel) and \
               ((fromModel is None) or (pair['onyx:conversionFrom'] == fromModel)):
                return True
        return False


EmotionConversionPlugin = EmotionConversion


class PostProcessing(Plugin):
    '''
    A plugin that converts the output of other plugins (post-processing).
    '''
    def check(self, request, plugins):
        '''Should this plugin be run for this request?'''
        return False


class Box(Analyser):
    '''
    Black box plugins delegate analysis to a function.
    The flow is like this:

    .. code-block::

                   entries --> to_features() --> predict_many() --> to_entry() --> entries'


    In other words: their ``to_features`` method converts a query (entry and a set of parameters)
    into the input to the `predict_one` method, which only uses an array of features.
    The ``to_entry`` method converts the results given by the box into an entry that senpy can
    handle.
    '''

    def to_features(self, entry, activity=None):
        '''Transforms a query (entry+param) into an input for the black box'''
        return entry

    def to_entry(self, features, entry=None, activity=None):
        '''Transforms the results of the black box into an entry'''
        return entry

    def predict_one(self, features, activity=None):
        raise NotImplementedError(
            'You should define the behavior of this plugin')

    def predict_many(self, features, activity=None):
        results = []
        for feat in features:
            results.append(self.predict_one(features=feat, activity=activity))
        return results

    def process_entry(self, entry, activity):
        for i in self.process_entries([entry], activity):
            yield i

    def process_entries(self, entries, activity):
        features = []
        for entry in entries:
            features.append(self.to_features(entry=entry, activity=activity))
        results = self.predict_many(features=features, activity=activity)

        for (result, entry) in zip(results, entries):
            yield self.to_entry(features=result, entry=entry, activity=activity)


class TextBox(Box):
    '''A black box plugin that takes only text as input'''

    def to_features(self, entry, activity):
        return [entry['nif:isString']]


class SentimentBox(TextBox, SentimentPlugin):
    '''
    A box plugin where the output is only a polarity label or a tuple (polarity, polarityValue)
    '''

    classes = ['marl:Positive', 'marl:Neutral', 'marl:Negative']
    binary = True

    def to_entry(self, features, entry, activity, **kwargs):

        if len(features) != len(self.classes):
            raise models.Error('The number of features ({}) does not match the classes '
                               '(plugin.classes ({})'.format(len(features), len(self.classes)))

        minValue = activity.param('marl:minPolarityValue', 0)
        maxValue = activity.param('marl:minPolarityValue', 1)
        activity['marl:minPolarityValue'] = minValue
        activity['marl:maxPolarityValue'] = maxValue

        for k, v in zip(self.classes, features):
            s = models.Sentiment()
            if self.binary:
                if not v:  # Carry on if the value is 0
                    continue
                s['marl:hasPolarity'] = k
            else:
                if v is not None:
                    s['marl:hasPolarity'] = k
                    nv = self.normalize(v, minValue, maxValue)
                    s['marl:polarityValue'] = nv
            s.prov(activity)

            entry.sentiments.append(s)

        return entry


class EmotionBox(TextBox, EmotionPlugin):
    '''
    A box plugin where the output is only an a tuple of emotion labels
    '''

    EMOTIONS = []
    with_intensity = True

    def to_entry(self, features, entry, activity, **kwargs):
        s = models.EmotionSet()

        if len(features) != len(self.EMOTIONS):
            raise Exception(('The number of classes in the plugin and the number of features '
                             'do not match'))

        for label, intensity in zip(self.EMOTIONS, features):
            e = models.Emotion(onyx__hasEmotionCategory=label)
            if self.with_intensity:
                e.onyx__hasEmotionIntensity = intensity
            s.onyx__hasEmotion.append(e)
        s.prov(activity)
        entry.emotions.append(s)
        return entry


class MappingMixin(object):
    @property
    def mappings(self):
        return self._mappings

    @mappings.setter
    def mappings(self, value):
        self._mappings = value

    def to_entry(self, features, entry, activity):
        features = list(features)
        for i, feat in enumerate(features):
            features[i] = self.mappings.get(feat,
                                            self.mappings.get('default',
                                                              feat))
        return super(MappingMixin, self).to_entry(features=features,
                                                  entry=entry,
                                                  activity=activity)


class ShelfMixin(object):
    @property
    def sh(self):
        if not hasattr(self, '_sh') or self._sh is None:
            self._sh = {}
            if os.path.isfile(self.shelf_file):
                try:
                    with self.open(self.shelf_file, 'rb') as p:
                        self._sh = pickle.load(p)
                except (IndexError, EOFError, pickle.UnpicklingError):
                    self.log.warning('Corrupted shelf file: {}'.format(
                        self.shelf_file))
                    if not self.get('force_shelf', False):
                        raise
        return self._sh

    @sh.deleter
    def sh(self):
        if os.path.isfile(self.shelf_file):
            os.remove(self.shelf_file)
            del self._sh
        self.save()

    @sh.setter
    def sh(self, value):
        self._sh = value

    @property
    def shelf_file(self):
        if not hasattr(self, '_shelf_file') or not self._shelf_file:
            self._shelf_file = os.path.join(self.data_folder, self.name + '.p')
        return self._shelf_file

    @shelf_file.setter
    def shelf_file(self, value):
        self._shelf_file = value

    def save(self):
        self.log.debug('Saving pickle')
        if hasattr(self, '_sh') and self._sh is not None:
            with self.open(self.shelf_file, 'wb') as f:
                pickle.dump(self._sh, f)


def pfilter(plugins, plugin_type=Analyser, **kwargs):
    """ Filter plugins by different criteria """
    if isinstance(plugins, models.Plugins):
        plugins = plugins.plugins
    elif isinstance(plugins, dict):
        plugins = plugins.values()
    logger.debug('#' * 100)
    logger.debug('plugin_type {}'.format(plugin_type))
    if plugin_type:
        if isinstance(plugin_type, PluginMeta):
            plugin_type = plugin_type.__name__
        try:
            plugin_type = plugin_type[0].upper() + plugin_type[1:]
            pclass = globals()[plugin_type]
            logger.debug('Class: {}'.format(pclass))
            candidates = filter(lambda x: isinstance(x, pclass), plugins)
        except KeyError:
            raise models.Error('{} is not a valid type'.format(plugin_type))
    else:
        candidates = plugins

    if 'name' in kwargs:
        kwargs['name'] = kwargs['name'].lower()

    logger.debug(candidates)

    def matches(plug):
        res = all(getattr(plug, k, None) == v for (k, v) in kwargs.items())
        logger.debug("matching {} with {}: {}".format(plug.name, kwargs, res))
        return res

    if kwargs:
        candidates = filter(matches, candidates)
    return candidates


def load_module(name, root=None):
    if root:
        sys.path.append(root)
    tmp = importlib.import_module(name)
    if root:
        sys.path.remove(root)
    return tmp


def _log_subprocess_output(process):
    for line in iter(process.stdout.readline, b''):
        logger.info('%s', line.decode())
    for line in iter(process.stderr.readline, b''):
        logger.error('%s', line.decode())


def missing_requirements(reqs):
    queue = []
    pool = multiprocessing.Pool(4)
    for req in reqs:
        res = pool.apply_async(pkg_resources.get_distribution, (req,))
        queue.append((req, res))
    missing = []
    for req, job in queue:
        try:
            job.get(1)
        except Exception:
            missing.append(req)
    return missing


def install_deps(*plugins):
    installed = False
    nltk_resources = set()
    requirements = []
    for info in plugins:
        requirements = info.get('requirements', [])
        if requirements:
            requirements += missing_requirements(requirements)
        nltk_resources |= set(info.get('nltk_resources', []))
    if requirements:
        logger.info('Installing requirements: ' + str(requirements))
        pip_args = [sys.executable, '-m', 'pip', 'install']
        for req in requirements:
            pip_args.append(req)
        process = subprocess.Popen(
            pip_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        _log_subprocess_output(process)
        exitcode = process.wait()
        installed = True
        if exitcode != 0:
            raise models.Error(
                "Dependencies not properly installed: {}".format(pip_args))
    installed |= download(list(nltk_resources))
    return installed


is_plugin_file = re.compile(r'.*\.senpy$|senpy_[a-zA-Z0-9_]+\.py$|'
                            '^(?!test_)[a-zA-Z0-9_]+_plugin.py$')


def find_plugins(folders):
    for search_folder in folders:
        for root, dirnames, filenames in os.walk(search_folder):
            # Do not look for plugins in hidden or special folders
            dirnames[:] = [d for d in dirnames if d[0] not in ['.', '_']]
            for filename in filter(is_plugin_file.match, filenames):
                fpath = os.path.join(root, filename)
                yield fpath


def from_path(fpath, install_on_fail=False, **kwargs):
    logger.debug("Loading plugin from {}".format(fpath))
    if fpath.endswith('.py'):
        # We asume root is the dir of the file, and module is the name of the file
        root = os.path.dirname(fpath)
        module = os.path.basename(fpath)[:-3]
        for instance in _from_module_name(module=module, root=root, **kwargs):
            yield instance
    else:
        info = parse_plugin_info(fpath)
        yield from_info(info, install_on_fail=install_on_fail, **kwargs)


def from_folder(folders, loader=from_path, **kwargs):
    plugins = []
    for fpath in find_plugins(folders):
        for plugin in loader(fpath, **kwargs):
            plugins.append(plugin)
    return plugins


def from_info(info, root=None, install_on_fail=True, **kwargs):
    if any(x not in info for x in ('module', )):
        raise ValueError('Plugin info is not valid: {}'.format(info))
    module = info["module"]

    if not root and '_path' in info:
        root = os.path.dirname(info['_path'])

    fun = partial(one_from_module, module, root=root, info=info, **kwargs)
    try:
        return fun()
    except (ImportError, LookupError):
        install_deps(info)
        return fun()


def parse_plugin_info(fpath):
    logger.debug("Parsing plugin info: {}".format(fpath))
    with open(fpath, 'r') as f:
        info = yaml.load(f, Loader=yaml.FullLoader)
    info['_path'] = fpath
    return info


def from_module(module, **kwargs):

    if inspect.ismodule(module):
        res = _from_loaded_module(module, **kwargs)
    else:
        res = _from_module_name(module, **kwargs)
    for p in res:
        yield p


def one_from_module(module, root, info, **kwargs):
    if '@type' in info:
        cls = PluginMeta.from_type(info['@type'])
        return cls(info=info, **kwargs)
    instance = next(
        from_module(module=module, root=root, info=info, **kwargs), None)
    if not instance:
        raise Exception("No valid plugin for: {}".format(module))
    return instance


def _classes_in_module(module):
    for _, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and inspect.getmodule(obj) == module:
            logger.debug(("Found plugin class:"
                          " {}@{}").format(obj, inspect.getmodule(obj)))
            yield obj


def _instances_in_module(module):
    for _, obj in inspect.getmembers(module):
        if isinstance(obj, Plugin) and inspect.getmodule(obj) == module:
            logger.debug(("Found plugin instance:"
                          " {}@{}").format(obj, inspect.getmodule(obj)))
            yield obj


def _from_module_name(module, root, info=None, **kwargs):
    module = load_module(module, root)
    for plugin in _from_loaded_module(
            module=module, root=root, info=info, **kwargs):
        yield plugin


def _from_loaded_module(module, info=None, **kwargs):
    for cls in _classes_in_module(module):
        yield cls(info=info, **kwargs)
    for instance in _instances_in_module(module):
        yield instance


cached_evs = {}


def evaluate(plugins, datasets, **kwargs):
    for plug in plugins:
        if not hasattr(plug, 'as_pipe'):
            raise models.Error('Plugin {} cannot be evaluated'.format(plug.name))

    if not isinstance(datasets, dict):
        datasets = gsitk_compat.prepare(datasets, download=True)

    tuples = list(product(plugins, datasets))
    missing = []
    for (p, d) in tuples:
        if (p.id, d) not in cached_evs:
            pipe = p.as_pipe()
            missing.append(gsitk_compat.EvalPipeline(pipe, d))
    if missing:
        ev = gsitk_compat.Eval(tuples=missing, datasets=datasets)
        ev.evaluate()
        results = ev.results
        new_ev = evaluations_to_JSONLD(results, **kwargs)
        for ev in new_ev:
            dataset = ev.evaluatesOn
            model = ev.evaluates
            cached_evs[(model, dataset)] = ev
    evaluations = []
    logger.debug('%s. Cached evs: %s', tuples, cached_evs)
    for (p, d) in tuples:
        logger.debug('Adding %s, %s', d, p)
        evaluations.append(cached_evs[(p.id, d)])
    return evaluations


def evaluations_to_JSONLD(results, flatten=False):
    '''
    Map the evaluation results to a JSONLD scheme
    '''

    evaluations = list()
    metric_names = ['accuracy', 'precision_macro', 'recall_macro',
                    'f1_macro', 'f1_weighted', 'f1_micro', 'f1_macro']

    for index, row in results.fillna('Not Available').iterrows():
        evaluation = models.Evaluation()
        if row.get('CV', True):
            evaluation['@type'] = ['StaticCV', 'Evaluation']
        evaluation.evaluatesOn = row['Dataset']
        evaluation.evaluates = row['Model'].rstrip('__' + row['Dataset'])
        i = 0
        if flatten:
            metric = models.Metric()
            for name in metric_names:
                metric[name] = row[name]
            evaluation.metrics.append(metric)
        else:
            # We should probably discontinue this representation
            for name in metric_names:
                metric = models.Metric()
                metric['@type'] = name.capitalize()
                metric.value = row[name]
                evaluation.metrics.append(metric)
                i += 1
        evaluations.append(evaluation)
    return evaluations


class ScikitWrapper(BaseEstimator, TransformerMixin):
    def __init__(self, plugin=None):
        self.plugin = plugin

    def fit(self, X=None, y=None):
        if self.plugin is not None and not self.plugin.is_activated:
            self.plugin.activate()
        return self

    def transform(self, X):
        return self.plugin.evaluate_func(X, None)

    def predict(self, X):
        return self.transform(X)

    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X)



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
from __future__ import print_function

import sys
from .models import Error
from .extensions import Senpy
from . import api


def argv_to_dict(argv):
    '''Turns parameters in the form of '--key value' into a dict {'key': 'value'}
    '''
    cli_dict = {}

    for i in range(len(argv)):
        if argv[i][0] == '-':
            key = argv[i].strip('-')
            value = argv[i + 1] if len(argv) > i + 1 else None
            if not value or value[0] == '-':
                cli_dict[key] = True
            else:
                cli_dict[key] = value
    return cli_dict


def main_function(argv):
    '''This is the method for unit testing
    '''
    params = api.parse_params(argv_to_dict(argv),
                              api.CLI_PARAMS,
                              api.API_PARAMS,
                              api.NIF_PARAMS)
    plugin_folder = params['plugin-folder']
    default_plugins = not params.get('no-default-plugins', False)
    sp = Senpy(default_plugins=default_plugins, plugin_folder=plugin_folder)
    request = api.parse_call(params)
    algos = sp.get_plugins(request.parameters.get('algorithm', None))
    if algos:
        for algo in algos:
            sp.activate_plugin(algo.name)
    else:
        sp.activate_all()
    res = sp.analyse(request)
    return res


def main():
    '''This method is the entrypoint for the CLI (as configured un setup.py)
    '''
    try:
        res = main_function(sys.argv[1:])
        print(res.serialize())
    except Error as err:
        print(err.serialize(), file=sys.stderr)
        sys.exit(2)


if __name__ == '__main__':
    main()



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Main class for Senpy.
It orchestrates plugin (de)activation and analysis.
"""
from future import standard_library
standard_library.install_aliases()

from . import plugins, api
from .models import Error, AggregatedEvaluation
from .plugins import AnalysisPlugin
from .blueprints import api_blueprint, demo_blueprint, ns_blueprint

from threading import Thread
from functools import partial
import os
import copy
import errno
import logging

from . import gsitk_compat

logger = logging.getLogger(__name__)


class Senpy(object):
    """ Default Senpy extension for Flask """

    def __init__(self,
                 app=None,
                 plugin_folder=".",
                 data_folder=None,
                 default_plugins=False):

        default_data = os.path.join(os.getcwd(), 'senpy_data')
        self.data_folder = data_folder or os.environ.get('SENPY_DATA', default_data)
        try:
            os.makedirs(self.data_folder)
        except OSError as e:
            if e.errno == errno.EEXIST:
                logger.debug('Data folder exists: {}'.format(self.data_folder))
            else:  # pragma: no cover
                raise

        self._default = None
        self._plugins = {}
        if plugin_folder:
            self.add_folder(plugin_folder)

        if default_plugins:
            self.add_folder('plugins', from_root=True)
        else:
            # Add only conversion plugins
            self.add_folder(os.path.join('plugins', 'postprocessing'),
                            from_root=True)
        self.app = app
        if app is not None:
            self.init_app(app)
        self._conversion_candidates = {}

    def init_app(self, app):
        """ Initialise a flask app to add plugins to its context """
        """
        Note: I'm not particularly fond of adding self.app and app.senpy, but
        I can't think of a better way to do it.
        """
        app.senpy = self
        # Use the newstyle teardown_appcontext if it's available,
        # otherwise fall back to the request context
        if hasattr(app, 'teardown_appcontext'):
            app.teardown_appcontext(self.teardown)
        else:  # pragma: no cover
            app.teardown_request(self.teardown)
        app.register_blueprint(api_blueprint, url_prefix="/api")
        app.register_blueprint(ns_blueprint, url_prefix="/ns")
        app.register_blueprint(demo_blueprint, url_prefix="/")

    def add_plugin(self, plugin):
        self._plugins[plugin.name.lower()] = plugin
        self._conversion_candidates = {}

    def delete_plugin(self, plugin):
        del self._plugins[plugin.name.lower()]

    def plugins(self, plugin_type=None, is_activated=True, **kwargs):
        """ Return the plugins registered for a given application. Filtered by criteria  """
        return sorted(plugins.pfilter(self._plugins,
                                      plugin_type=plugin_type,
                                      is_activated=is_activated,
                                      **kwargs),
                      key=lambda x: x.id)

    def get_plugin(self, name, default=None):
        if name == 'default':
            return self.default_plugin
        elif name == 'conversion':
            return None

        if name.lower() in self._plugins:
            return self._plugins[name.lower()]

        results = self.plugins(id='endpoint:plugins/{}'.format(name.lower()),
                               plugin_type=None)
        if results:
            return results[0]

        results = self.plugins(id=name,
                               plugin_type=None)
        if results:
            return results[0]

        msg = ("Plugin not found: '{}'\n"
               "Make sure it is ACTIVATED\n"
               "Valid algorithms: {}").format(name,
                                              self._plugins.keys())
        raise Error(message=msg, status=404)

    def get_plugins(self, name):
        try:
            name = name.split(',')
        except AttributeError:
            pass  # Assume it is a tuple or a list
        return tuple(self.get_plugin(n) for n in name)

    def analysis_plugins(self, **kwargs):
        """ Return only the analysis plugins that are active"""
        candidates = self.plugins(**kwargs)
        return list(plugins.pfilter(candidates, plugin_type=AnalysisPlugin))

    def add_folder(self, folder, from_root=False):
        """ Find plugins in this folder and add them to this instance """
        if from_root:
            folder = os.path.join(os.path.dirname(__file__), folder)
        logger.debug("Adding folder: %s", folder)
        if os.path.isdir(folder):
            new_plugins = plugins.from_folder([folder],
                                              data_folder=self.data_folder)
            for plugin in new_plugins:
                self.add_plugin(plugin)
        else:
            raise AttributeError("Not a folder or does not exist: %s", folder)

    def _process(self, req, pending, done=None):
        """
        Recursively process the entries with the first plugin in the list, and pass the results
        to the rest of the plugins.
        """
        done = done or []
        if not pending:
            return req

        analysis = pending[0]
        results = analysis.run(req)
        results.activities.append(analysis)
        done += analysis
        return self._process(results, pending[1:], done)

    def install_deps(self):
        logger.info('Installing dependencies')
        # If a plugin is activated, its dependencies should already be installed
        # Otherwise, it would've failed to activate.
        plugins.install_deps(*self.plugins(is_activated=False))

    def analyse(self, request, analyses=None):
        """
        Main method that analyses a request, either from CLI or HTTP.
        It takes a processed request, provided by the user, as returned
        by api.parse_call().
        """
        if not self.plugins():
            raise Error(
                status=404,
                message=("No plugins found."
                         " Please install one."))
        if analyses is None:
            plugins = self.get_plugins(request.parameters['algorithm'])
            analyses = api.parse_analyses(request.parameters, plugins)
        logger.debug("analysing request: {}".format(request))
        results = self._process(request, analyses)
        logger.debug("Got analysis result: {}".format(results))
        results = self.postprocess(results, analyses)
        logger.debug("Returning post-processed result: {}".format(results))
        return results

    def convert_emotions(self, resp, analyses):
        """
        Conversion of all emotions in a response **in place**.
        In addition to converting from one model to another, it has
        to include the conversion plugin to the analysis list.
        Needless to say, this is far from an elegant solution, but it works.
        @todo refactor and clean up
        """

        logger.debug("Converting emotions")
        if 'parameters' not in resp:
            logger.debug("NO PARAMETERS")
            return resp

        params = resp['parameters']
        toModel = params.get('emotion-model', None)
        if not toModel:
            logger.debug("NO tomodel PARAMETER")
            return resp

        logger.debug('Asked for model: {}'.format(toModel))
        output = params.get('conversion', None)

        newentries = []
        done = []
        for i in resp.entries:

            if output == "full":
                newemotions = copy.deepcopy(i.emotions)
            else:
                newemotions = []
            for j in i.emotions:
                activity = j['prov:wasGeneratedBy']
                act = resp.activity(activity)
                if not act:
                    raise Error('Could not find the emotion model for {}'.format(activity))
                fromModel = act.plugin['onyx:usesEmotionModel']
                if toModel == fromModel:
                    continue
                candidate = self._conversion_candidate(fromModel, toModel)
                if not candidate:
                    e = Error(('No conversion plugin found for: '
                              '{} -> {}'.format(fromModel, toModel)),
                              status=404)
                    e.original_response = resp
                    e.parameters = params
                    raise e

                analysis = candidate.activity(params)
                done.append(analysis)
                for k in candidate.convert(j, fromModel, toModel, params):
                    k.prov__wasGeneratedBy = analysis.id
                    if output == 'nested':
                        k.prov__wasDerivedFrom = j
                    newemotions.append(k)
            i.emotions = newemotions
            newentries.append(i)
        resp.entries = newentries
        return resp

    def _conversion_candidate(self, fromModel, toModel):
        if not self._conversion_candidates:
            candidates = {}
            for conv in self.plugins(plugin_type=plugins.EmotionConversion):
                for pair in conv.onyx__doesConversion:
                    logging.debug(pair)
                    key = (pair['onyx:conversionFrom'], pair['onyx:conversionTo'])
                    if key not in candidates:
                        candidates[key] = []
                    candidates[key].append(conv)
            self._conversion_candidates = candidates

        key = (fromModel, toModel)
        if key not in self._conversion_candidates:
            return None
        return self._conversion_candidates[key][0]

    def postprocess(self, response, analyses):
        '''
        Transform the results from the analysis plugins.
        It has some pre-defined post-processing like emotion conversion,
        and it also allows plugins to auto-select themselves.
        '''

        response = self.convert_emotions(response, analyses)

        for plug in self.plugins(plugin_type=plugins.PostProcessing):
            if plug.check(response, response.activities):
                activity = plug.activity(response.parameters)
                response = plug.process(response, activity)
        return response

    def _get_datasets(self, request):
        datasets_name = request.parameters.get('dataset', None).split(',')
        for dataset in datasets_name:
            if dataset not in gsitk_compat.datasets:
                logger.debug(("The dataset '{}' is not valid\n"
                              "Valid datasets: {}").format(
                                  dataset, gsitk_compat.datasets.keys()))
                raise Error(
                    status=404,
                    message="The dataset '{}' is not valid".format(dataset))
        return datasets_name

    def evaluate(self, params):
        logger.debug("evaluating request: {}".format(params))
        results = AggregatedEvaluation()
        results.parameters = params
        datasets = self._get_datasets(results)
        plugs = []
        for plugname in params['algorithm']:
            plugs = self.get_plugins(plugname)
        for plug in plugs:
            if not isinstance(plug, plugins.Evaluable):
                raise Exception('Plugin {} can not be evaluated', plug.id)

        for eval in plugins.evaluate(plugs, datasets):
            results.evaluations.append(eval)
        if 'with-parameters' not in results.parameters:
            del results.parameters
        logger.debug("Returning evaluation result: {}".format(results))
        return results

    @property
    def default_plugin(self):
        if not self._default or not self._default.is_activated:
            candidates = self.analysis_plugins()
            if len(candidates) > 0:
                self._default = candidates[0]
            else:
                self._default = None
            logger.debug("Default: {}".format(self._default))
        return self._default

    @default_plugin.setter
    def default_plugin(self, value):
        if isinstance(value, plugins.Plugin):
            if not value.is_activated:
                raise AttributeError('The default plugin has to be activated.')
            self._default = value

        else:
            self._default = self._plugins[value.lower()]

    def activate_all(self, sync=True, allow_fail=False):
        ps = []
        for plug in self._plugins.keys():
            try:
                self.activate_plugin(plug, sync=sync)
            except Exception as ex:
                if not allow_fail:
                    raise
                logger.error('Could not activate {}: {}'.format(plug, ex))
        return ps

    def deactivate_all(self, sync=True):
        ps = []
        for plug in self._plugins.keys():
            ps.append(self.deactivate_plugin(plug, sync=sync))
        return ps

    def _activate(self, plugin):
        success = False
        with plugin._lock:
            if plugin.is_activated:
                return
            plugin._activate()
            msg = "Plugin activated: {}".format(plugin.name)
            logger.info(msg)
            success = plugin.is_activated
        return success

    def activate_plugin(self, plugin_name, sync=True):
        plugin_name = plugin_name.lower()
        if plugin_name not in self._plugins:
            raise Error(
                message="Plugin not found: {}".format(plugin_name), status=404)
        plugin = self._plugins[plugin_name]

        logger.info("Activating plugin: {}".format(plugin.name))

        if sync or not getattr(plugin, 'async', True) or getattr(
                plugin, 'sync', False):
            return self._activate(plugin)
        else:
            th = Thread(target=partial(self._activate, plugin))
            th.start()
            return th

    def _deactivate(self, plugin):
        with plugin._lock:
            if not plugin.is_activated:
                return
            plugin._deactivate()
            logger.info("Plugin deactivated: {}".format(plugin.name))

    def deactivate_plugin(self, plugin_name, sync=True):
        plugin_name = plugin_name.lower()
        if plugin_name not in self._plugins:
            raise Error(
                message="Plugin not found: {}".format(plugin_name), status=404)
        plugin = self._plugins[plugin_name]

        if sync or not getattr(plugin, 'async', True) or not getattr(
                plugin, 'sync', False):
            plugin._deactivate()
        else:
            th = Thread(target=plugin.deactivate)
            th.start()
            return th

    def teardown(self, exception):
        pass



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from past.builtins import basestring

import os
import responses as requestmock

from .models import BaseModel


MOCK_REQUESTS = os.environ.get('MOCK_REQUESTS', '').lower() in ['no', 'false']


def patch_all_requests(responses):

    patched = requestmock.RequestsMock()

    for response in responses or []:
        args = response.copy()
        if 'json' in args and isinstance(args['json'], BaseModel):
            args['json'] = args['json'].jsonld()
        args['method'] = getattr(requestmock, args.get('method', 'GET'))
        patched.add(**args)
    return patched


def patch_requests(url, response, method='GET', status=200):
    args = {'url': url, 'method': method, 'status': status}
    if isinstance(response, basestring):
        args['body'] = response
    else:
        args['json'] = response
    return patch_all_requests([args])



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Blueprints for Senpy
"""
from flask import (Blueprint, request, current_app, render_template, url_for,
                   jsonify, redirect)
from .models import Error, Response, Help, Plugins, read_schema, dump_schema, Datasets
from . import api
from .version import __version__
from functools import wraps

from .gsitk_compat import GSITK_AVAILABLE, datasets

import logging
import json
import base64

logger = logging.getLogger(__name__)

api_blueprint = Blueprint("api", __name__)
demo_blueprint = Blueprint("demo", __name__, template_folder='templates')
ns_blueprint = Blueprint("ns", __name__)

_mimetypes_r = {'json-ld': ['application/ld+json'],
                'turtle': ['text/turtle'],
                'ntriples': ['application/n-triples'],
                'text': ['text/plain']}

MIMETYPES = {}

for k, vs in _mimetypes_r.items():
    for v in vs:
        if v in MIMETYPES:
            raise Exception('MIMETYPE {} specified for two formats: {} and {}'.format(v,
                                                                                      v,
                                                                                      MIMETYPES[v]))
        MIMETYPES[v] = k

DEFAULT_MIMETYPE = 'application/ld+json'
DEFAULT_FORMAT = 'json-ld'


def get_params(req):
    if req.method == 'POST':
        indict = req.form.to_dict(flat=True)
    elif req.method == 'GET':
        indict = req.args.to_dict(flat=True)
    else:
        raise Error(message="Invalid data")
    return indict


def encode_url(url=None):
    code = ''
    if not url:
        url = request.parameters.get('prefix', request.full_path[1:] + '#')
    return code or base64.urlsafe_b64encode(url.encode()).decode()


def url_for_code(code, base=None):
    # if base:
    #     return base + code
    # return url_for('api.decode', code=code, _external=True)
    # This was producing unique yet very long URIs, which wasn't ideal for visualization.
    return 'http://senpy.invalid/'


def decoded_url(code, base=None):
    path = base64.urlsafe_b64decode(code.encode()).decode()
    if path[:4] == 'http':
        return path
    base = base or request.url_root
    return base + path


@demo_blueprint.route('/')
def index():
    # ev = str(get_params(request).get('evaluation', True))
    # evaluation_enabled = ev.lower() not in ['false', 'no', 'none']
    evaluation_enabled = GSITK_AVAILABLE

    return render_template("index.html",
                           evaluation=evaluation_enabled,
                           version=__version__)


@api_blueprint.route('/contexts/<code>')
def context(code=''):
    context = Response._context
    context['@base'] = url_for('api.decode', code=code, _external=True)
    context['endpoint'] = url_for('api.api_root', _external=True)
    return jsonify({"@context": context})


@api_blueprint.route('/d/<code>')
def decode(code):
    try:
        return redirect(decoded_url(code))
    except Exception:
        return Error('invalid URL').flask()


@ns_blueprint.route('/')  # noqa: F811
def index():
    context = Response._context.copy()
    context['endpoint'] = url_for('api.api_root', _external=True)
    return jsonify({"@context": context})


@api_blueprint.route('/schemas/<schema>')
def schema(schema="definitions"):
    try:
        return dump_schema(read_schema(schema))
    except Exception as ex:  # Should be FileNotFoundError, but it's missing from py2
        return Error(message="Schema not found: {}".format(ex), status=404).flask()


def basic_api(f):
    default_params = {
        'in-headers': False,
        'expanded-jsonld': False,
        'outformat': None,
        'with-parameters': True,
    }

    @wraps(f)
    def decorated_function(*args, **kwargs):
        raw_params = get_params(request)
        # logger.info('Getting request: {}'.format(raw_params))
        logger.debug('Getting request. Params: {}'.format(raw_params))
        headers = {'X-ORIGINAL-PARAMS': json.dumps(raw_params)}
        params = default_params

        mime = request.accept_mimetypes\
                      .best_match(MIMETYPES.keys(),
                                  DEFAULT_MIMETYPE)

        mimeformat = MIMETYPES.get(mime, DEFAULT_FORMAT)
        outformat = mimeformat

        try:
            params = api.parse_params(raw_params, api.WEB_PARAMS, api.API_PARAMS)
            outformat = params.get('outformat', mimeformat)
            if hasattr(request, 'parameters'):
                request.parameters.update(params)
            else:
                request.parameters = params
            response = f(*args, **kwargs)

            if 'parameters' in response and not params['with-parameters']:
                del response.parameters

            logger.debug('Response: {}'.format(response))

            prefix = params.get('prefix')
            code = encode_url(prefix)

            return response.flask(
                in_headers=params['in-headers'],
                headers=headers,
                prefix=prefix or url_for_code(code),
                base=prefix,
                context_uri=url_for('api.context',
                                    code=code,
                                    _external=True),
                outformat=outformat,
                expanded=params['expanded-jsonld'],
                template=params.get('template'),
                verbose=params['verbose'],
                aliases=params['aliases'],
                fields=params.get('fields'))

        except (Exception) as ex:
            if current_app.debug or current_app.config['TESTING']:
                raise
            if not isinstance(ex, Error):
                msg = "{}".format(ex)
                ex = Error(message=msg, status=500)
            response = ex
            response.parameters = raw_params
            logger.exception(ex)
            return response.flask(
                outformat=outformat,
                expanded=params['expanded-jsonld'],
                verbose=params.get('verbose', True),
            )

    return decorated_function


@api_blueprint.route('/', defaults={'plugins': None}, methods=['POST', 'GET'], strict_slashes=False)
@api_blueprint.route('/<path:plugins>', methods=['POST', 'GET'], strict_slashes=False)
@basic_api
def api_root(plugins):
    if plugins:
        if request.parameters['algorithm'] != api.API_PARAMS['algorithm']['default']:
            raise Error('You cannot specify the algorithm with a parameter and a URL variable.'
                        ' Please, remove one of them')
        plugins = plugins.replace('+', ',').replace('/', ',')
        plugins = api.processors['string_to_tuple'](plugins)
    else:
        plugins = request.parameters['algorithm']

    print(plugins)

    sp = current_app.senpy
    plugins = sp.get_plugins(plugins)

    if request.parameters['help']:
        apis = [api.WEB_PARAMS, api.API_PARAMS, api.NIF_PARAMS]
        # Verbose is set to False as default, but we want it to default to
        # True for help. This checks the original value, to make sure it wasn't
        # set by default.
        if not request.parameters['verbose'] and get_params(request).get('verbose'):
            apis = []
        if request.parameters['algorithm'] == ['default', ]:
            plugins = []
        allparameters = api.get_all_params(plugins, *apis)
        response = Help(valid_parameters=allparameters)
        return response
    req = api.parse_call(request.parameters)
    analyses = api.parse_analyses(req.parameters, plugins)
    results = current_app.senpy.analyse(req, analyses)
    return results


@api_blueprint.route('/evaluate', methods=['POST', 'GET'], strict_slashes=False)
@basic_api
def evaluate():
    if request.parameters['help']:
        dic = dict(api.EVAL_PARAMS)
        response = Help(parameters=dic)
        return response
    else:
        params = api.parse_params(request.parameters, api.EVAL_PARAMS)
        response = current_app.senpy.evaluate(params)
        return response


@api_blueprint.route('/plugins', methods=['POST', 'GET'], strict_slashes=False)
@basic_api
def plugins():
    sp = current_app.senpy
    params = api.parse_params(request.parameters, api.PLUGINS_PARAMS)
    ptype = params.get('plugin-type')
    plugins = list(sp.analysis_plugins(plugin_type=ptype))
    dic = Plugins(plugins=plugins)
    return dic


@api_blueprint.route('/plugins/<plugin>', methods=['POST', 'GET'], strict_slashes=False)
@basic_api
def plugin(plugin):
    sp = current_app.senpy
    return sp.get_plugin(plugin)


@api_blueprint.route('/datasets', methods=['POST', 'GET'], strict_slashes=False)
@basic_api
def get_datasets():
    dic = Datasets(datasets=list(datasets.values()))
    return dic



#!/usr/bin/python
# -*- coding: utf-8 -*-
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""
Senpy is a modular sentiment analysis server. This script runs an instance of
the server.

"""

from flask import Flask
from senpy.extensions import Senpy
from senpy.utils import easy_test

import logging
import os
import sys
import argparse
import senpy

SERVER_PORT = os.environ.get("PORT", 5000)


def main():
    parser = argparse.ArgumentParser(description='Run a Senpy server')
    parser.add_argument(
        '--level',
        '-l',
        metavar='logging_level',
        type=str,
        default="INFO",
        help='Logging level')
    parser.add_argument(
        '--no-proxy-fix',
        action='store_true',
        default=False,
        help='Do not assume senpy will be running behind a proxy (e.g., nginx)')
    parser.add_argument(
        '--log-format',
        metavar='log_format',
        type=str,
        default='%(asctime)s %(levelname)-10s %(name)-30s \t %(message)s',
        help='Logging format')
    parser.add_argument(
        '--debug',
        '-d',
        action='store_true',
        default=False,
        help='Run the application in debug mode')
    parser.add_argument(
        '--no-default-plugins',
        action='store_true',
        default=False,
        help='Do not load the default plugins')
    parser.add_argument(
        '--host',
        type=str,
        default="0.0.0.0",
        help='Use 0.0.0.0 to accept requests from any host.')
    parser.add_argument(
        '--port',
        '-p',
        type=int,
        default=SERVER_PORT,
        help='Port to listen on.')
    parser.add_argument(
        '--plugins-folder',
        '-f',
        type=str,
        action='append',
        help='Where to look for plugins.')
    parser.add_argument(
        '--only-install',
        '-i',
        action='store_true',
        default=False,
        help='Do not run a server, only install plugin dependencies')
    parser.add_argument(
        '--only-test',
        action='store_true',
        default=False,
        help='Do not run a server, just test all plugins')
    parser.add_argument(
        '--test',
        '-t',
        action='store_true',
        default=False,
        help='Test all plugins before launching the server')
    parser.add_argument(
        '--only-list',
        '--list',
        action='store_true',
        default=False,
        help='Do not run a server, only list plugins found')
    parser.add_argument(
        '--data-folder',
        '--data',
        type=str,
        default=None,
        help='Where to look for data. It be set with the SENPY_DATA environment variable as well.')
    parser.add_argument(
        '--no-threaded',
        action='store_true',
        default=False,
        help='Run a single-threaded server')
    parser.add_argument(
        '--no-deps',
        '-n',
        action='store_true',
        default=False,
        help='Skip installing dependencies')
    parser.add_argument(
        '--version',
        '-v',
        action='store_true',
        default=False,
        help='Output the senpy version and exit')
    parser.add_argument(
        '--allow-fail',
        '--fail',
        action='store_true',
        default=False,
        help='Do not exit if some plugins fail to activate')
    parser.add_argument(
        '--enable-cors',
        '--cors',
        action='store_true',
        default=False,
        help='Enable CORS for all domains (requires flask-cors to be installed)')
    args = parser.parse_args()
    print('Senpy version {}'.format(senpy.__version__))
    print(sys.version)
    if args.version:
        exit(1)
    rl = logging.getLogger()
    rl.setLevel(getattr(logging, args.level))
    logger_handler = rl.handlers[0]

    # First, generic formatter:
    logger_handler.setFormatter(logging.Formatter(args.log_format))

    app = Flask(__name__)
    app.debug = args.debug

    sp = Senpy(app,
               plugin_folder=None,
               default_plugins=not args.no_default_plugins,
               data_folder=args.data_folder)
    folders = list(args.plugins_folder) if args.plugins_folder else []
    if not folders:
        folders.append(".")
    for p in folders:
        sp.add_folder(p)

    plugins = sp.plugins(plugin_type=None, is_activated=False)
    maxname = max(len(x.name) for x in plugins)
    maxversion = max(len(str(x.version)) for x in plugins)
    print('Found {} plugins:'.format(len(plugins)))
    for plugin in plugins:
        import inspect
        fpath = inspect.getfile(plugin.__class__)
        print('\t{: <{maxname}} @ {: <{maxversion}} -> {}'.format(plugin.name,
                                                                  plugin.version,
                                                                  fpath,
                                                                  maxname=maxname,
                                                                  maxversion=maxversion))
    if args.only_list:
        return
    if not args.no_deps:
        sp.install_deps()
    if args.only_install:
        return
    sp.activate_all(allow_fail=args.allow_fail)
    if args.test or args.only_test:
        easy_test(sp.plugins(), debug=args.debug)
        if args.only_test:
            return
    print('Senpy version {}'.format(senpy.__version__))
    print('Server running on port %s:%d. Ctrl+C to quit' % (args.host,
                                                            args.port))
    if args.enable_cors:
        from flask_cors import CORS
        CORS(app)

    if not args.no_proxy_fix:
        from werkzeug.middleware.proxy_fix import ProxyFix
        app.wsgi_app = ProxyFix(app.wsgi_app)

    try:
        app.run(args.host,
                args.port,
                threaded=not args.no_threaded,
                debug=app.debug)
    except KeyboardInterrupt:
        print('Bye!')
    sp.deactivate_all()


if __name__ == '__main__':
    main()



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from senpy.plugins import Transformation
from senpy.models import Entry
from nltk.tokenize.punkt import PunktSentenceTokenizer
from nltk.tokenize.simple import LineTokenizer


class Split(Transformation):
    '''
    A plugin that chunks input text, into paragraphs or sentences.

    It does not provide any sort of annotation, and it is meant to precede
    other annotation plugins, when the annotation of individual sentences
    (or paragraphs) is required.
    '''

    author = ["@militarpancho", '@balkian']
    version = '0.3'
    url = "https://github.com/gsi-upm/senpy"
    nltk_resources = ['punkt']

    extra_params = {
        'delimiter': {
            'description': 'Split text into paragraphs or sentences.',
            'aliases': ['type', 't'],
            'required': False,
            'default': 'sentence',
            'options': ['sentence', 'paragraph']
        },
    }

    def analyse_entry(self, entry, activity):
        yield entry
        chunker_type = activity.params["delimiter"]
        original_text = entry['nif:isString']
        if chunker_type == "sentence":
            tokenizer = PunktSentenceTokenizer()
        if chunker_type == "paragraph":
            tokenizer = LineTokenizer()
        chars = list(tokenizer.span_tokenize(original_text))
        if len(chars) == 1:
            # This sentence was already split
            return
        for i, chunk in enumerate(chars):
            start, end = chunk
            e = Entry()
            e['nif:isString'] = original_text[start:end]
            if entry.id:
                e.id = entry.id + "#char={},{}".format(start, end)
            yield e

    test_cases = [
        {
            'entry': {
                'nif:isString': 'Hello. World.'
            },
            'params': {
                'delimiter': 'sentence',
            },
            'expected': [
                {
                    'nif:isString': 'Hello.'
                },
                {
                    'nif:isString': 'World.'
                }
            ]
        },
        {
            'entry': {
                "@id": ":test",
                'nif:isString': 'Hello\nWorld'
            },
            'params': {
                'delimiter': 'paragraph',
            },
            'expected': [
                {
                    "@id": ":test#char=0,5",
                    'nif:isString': 'Hello'
                },
                {
                    "@id": ":test#char=6,11",
                    'nif:isString': 'World'
                }
            ]
        }
    ]



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from senpy import PostProcessing, easy_test


class MaxEmotion(PostProcessing):
    '''Plugin to extract the emotion with highest value from an EmotionSet'''
    author = '@dsuarezsouto'
    version = '0.1'

    def process_entry(self, entry, activity):
        if len(entry.emotions) < 1:
            yield entry
            return

        set_emotions = entry.emotions[0]['onyx:hasEmotion']

        # If there is only one emotion, do not modify it
        if len(set_emotions) < 2:
            yield entry
            return

        max_emotion = set_emotions[0]

        # Extract max emotion from the set emotions (emotion with highest intensity)
        for tmp_emotion in set_emotions:
            if tmp_emotion['onyx:hasEmotionIntensity'] > max_emotion[
                    'onyx:hasEmotionIntensity']:
                max_emotion = tmp_emotion

        if max_emotion['onyx:hasEmotionIntensity'] == 0:
            max_emotion['onyx:hasEmotionCategory'] = "neutral"
            max_emotion['onyx:hasEmotionIntensity'] = 1.0

        entry.emotions[0]['onyx:hasEmotion'] = [max_emotion]

        entry.emotions[0]['prov:wasGeneratedBy'] = activity.id
        yield entry

    def check(self, request, plugins):
        return 'maxemotion' in request.parameters and self not in plugins

    # Test Cases:
    #   1 Normal Situation.
    #   2 Case to return a Neutral Emotion.
    test_cases = [
        {
            "name": "If there are several emotions within an emotion set, reduce it to one.",
            "entry": {
                "@type":
                "entry",
                "onyx:hasEmotionSet": [
                    {
                        "@id":
                        "Emotions0",
                        "@type":
                        "emotionSet",
                        "onyx:hasEmotion": [
                            {
                                "@id": "_:Emotion_1538121033.74",
                                "@type": "emotion",
                                "onyx:hasEmotionCategory": "anger",
                                "onyx:hasEmotionIntensity": 0
                            },
                            {
                                "@id": "_:Emotion_1538121033.74",
                                "@type": "emotion",
                                "onyx:hasEmotionCategory": "joy",
                                "onyx:hasEmotionIntensity": 0.3333333333333333
                            },
                            {
                                "@id": "_:Emotion_1538121033.74",
                                "@type": "emotion",
                                "onyx:hasEmotionCategory": "negative-fear",
                                "onyx:hasEmotionIntensity": 0
                            },
                            {
                                "@id": "_:Emotion_1538121033.74",
                                "@type": "emotion",
                                "onyx:hasEmotionCategory": "sadness",
                                "onyx:hasEmotionIntensity": 0
                            },
                            {
                                "@id": "_:Emotion_1538121033.74",
                                "@type": "emotion",
                                "onyx:hasEmotionCategory": "disgust",
                                "onyx:hasEmotionIntensity": 0
                            }
                        ]
                    }
                ],
                "nif:isString":
                "Test"
            },
            'expected': {
                "@type":
                "entry",
                "onyx:hasEmotionSet": [
                    {
                        "@id":
                        "Emotions0",
                        "@type":
                        "emotionSet",
                        "onyx:hasEmotion": [
                            {
                                "@id": "_:Emotion_1538121033.74",
                                "@type": "emotion",
                                "onyx:hasEmotionCategory": "joy",
                                "onyx:hasEmotionIntensity": 0.3333333333333333
                            }
                        ]
                    }
                ],
                "nif:isString":
                "Test"
            }
        },
        {
            "name":
            "If the maximum emotion has an intensity of 0, return a neutral emotion.",
            "entry": {
                "@type":
                "entry",
                "onyx:hasEmotionSet": [{
                    "@id":
                    "Emotions0",
                    "@type":
                    "emotionSet",
                    "onyx:hasEmotion": [
                        {
                            "@id": "_:Emotion_1538121033.74",
                            "@type": "emotion",
                            "onyx:hasEmotionCategory": "anger",
                            "onyx:hasEmotionIntensity": 0
                        },
                        {
                            "@id": "_:Emotion_1538121033.74",
                            "@type": "emotion",
                            "onyx:hasEmotionCategory": "joy",
                            "onyx:hasEmotionIntensity": 0
                        },
                        {
                            "@id":
                            "_:Emotion_1538121033.74",
                            "@type":
                            "emotion",
                            "onyx:hasEmotionCategory":
                            "negative-fear",
                            "onyx:hasEmotionIntensity":
                            0
                        },
                        {
                            "@id": "_:Emotion_1538121033.74",
                            "@type": "emotion",
                            "onyx:hasEmotionCategory":
                            "sadness",
                            "onyx:hasEmotionIntensity": 0
                        },
                        {
                            "@id": "_:Emotion_1538121033.74",
                            "@type": "emotion",
                            "onyx:hasEmotionCategory":
                            "disgust",
                            "onyx:hasEmotionIntensity": 0
                        }]
                }],
                "nif:isString":
                "Test"
            },
            'expected': {
                "@type":
                "entry",
                "onyx:hasEmotionSet": [{
                    "@id":
                    "Emotions0",
                    "@type":
                    "emotionSet",
                    "onyx:hasEmotion": [{
                        "@id": "_:Emotion_1538121033.74",
                        "@type": "emotion",
                        "onyx:hasEmotionCategory": "neutral",
                        "onyx:hasEmotionIntensity": 1
                    }]
                }],
                "nif:isString":
                "Test"
            }
        }
    ]


if __name__ == '__main__':
    easy_test()



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
from . import models, __version__
from collections.abc import MutableMapping
import pprint
import pdb

import logging
logger = logging.getLogger(__name__)

# MutableMapping should be enough, but it causes problems with py2
DICTCLASSES = (MutableMapping, dict, models.BaseModel)


def check_template(indict, template):
    if isinstance(template, DICTCLASSES) and isinstance(indict, DICTCLASSES):
        for k, v in template.items():
            if k not in indict:
                raise models.Error('{} not in {}'.format(k, indict))
            check_template(indict[k], v)
    elif isinstance(template, list) and isinstance(indict, list):
        for e in template:
            for i in indict:
                try:
                    check_template(i, e)
                    break
                except models.Error as ex:
                    # raise
                    continue
            else:
                raise models.Error(('Element not found.'
                                   '\nExpected: {}\nIn: {}').format(pprint.pformat(e),
                                                                    pprint.pformat(indict)))
    else:
        if indict != template:
            raise models.Error(('Differences found.\n'
                                '\tExpected: {}\n'
                                '\tFound: {}').format(pprint.pformat(template),
                                                      pprint.pformat(indict)))


def convert_dictionary(original, mappings):
    result = {}
    for key, value in original.items():
        if key in mappings:
            key = mappings[key]
        result[key] = value
    return result


def easy_load(app=None, plugin_list=None, plugin_folder=None, **kwargs):
    '''
    Run a server with a specific plugin.
    '''

    from flask import Flask
    from .extensions import Senpy

    if not app:
        app = Flask(__name__)
    sp = Senpy(app, plugin_folder=plugin_folder, **kwargs)
    if not plugin_list:
        from . import plugins
        import __main__
        plugin_list = plugins.from_module(__main__)
    for plugin in plugin_list:
        sp.add_plugin(plugin)
    sp.install_deps()
    sp.activate_all()
    return sp, app


def easy_test(plugin_list=None, debug=True):
    logger.setLevel(logging.DEBUG)
    logging.getLogger().setLevel(logging.INFO)
    try:
        if not plugin_list:
            import __main__
            logger.info('Loading classes from {}'.format(__main__))
            from . import plugins
            plugin_list = plugins.from_module(__main__)
        plugin_list = list(plugin_list)
        for plug in plugin_list:
            plug.test()
            plug.log.info('My tests passed!')
        logger.info('All tests passed for {} plugins!'.format(len(plugin_list)))
    except Exception:
        if not debug:
            raise
        pdb.post_mortem()


def easy(host='0.0.0.0', port=5000, debug=False, **kwargs):
    '''
    Run a server with a specific plugin.
    '''
    logging.getLogger().setLevel(logging.DEBUG)
    logging.getLogger('senpy').setLevel(logging.INFO)
    sp, app = easy_load(**kwargs)
    easy_test(sp.plugins())
    app.debug = debug
    import time
    logger.info(time.time())
    logger.info('Senpy version {}'.format(__version__))
    logger.info('Server running on port %s:%d. Ctrl+C to quit' % (host,
                                                                  port))
    app.debug = debug
    app.run(host,
            port,
            debug=app.debug)



'''
Copyright 2014 GSI DIT UPM

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from setuptools import setup
from os import path
import io

try:
    with io.open('senpy/VERSION') as f:
        __version__ = f.read().strip()
        assert __version__
except IOError:  # pragma: no cover
    print('Installing a development version of senpy. Proceed with caution!')
    __version__ = 'devel'


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    with io.open(filename, 'r') as f:
        lineiter = list(line.strip() for line in f)
    return [line for line in lineiter if line and not line.startswith("#")]


this_directory = path.abspath(path.dirname(__file__))
with io.open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


install_reqs = parse_requirements("requirements.txt")
test_reqs = parse_requirements("test-requirements.txt")
extra_reqs = parse_requirements("extra-requirements.txt")


setup(
    name='senpy',
    python_requires='>3.6',
    packages=['senpy'],  # this must be the same as the name above
    version=__version__,
    description=('A sentiment analysis server implementation. '
                 'Designed to be extensible, so new algorithms '
                 'and sources can be used.'),
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author='J. Fernando Sanchez',
    author_email='balkian@gmail.com',
    url='https://github.com/gsi-upm/senpy',  # use the URL to the github repo
    download_url='https://github.com/gsi-upm/senpy/archive/{}.tar.gz'.format(
        __version__),
    keywords=['eurosentiment', 'sentiment', 'emotions', 'nif'],
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    install_requires=install_reqs,
    tests_require=test_reqs,
    setup_requires=['pytest-runner', ],
    extras_require={
        'evaluation': extra_reqs,
        'extras': extra_reqs
    },
    include_package_data=True,
    entry_points={
        'console_scripts':
        ['senpy = senpy.__main__:main', 'senpy-cli = senpy.cli:main']
    })



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
'''
Meta-programming for the models.
'''
import os
import json
import jsonschema
import inspect
import copy

from abc import ABCMeta
from collections import namedtuple
from collections.abc import MutableMapping


class BaseMeta(ABCMeta):
    '''
    Metaclass for models. It extracts the default values for the fields in
    the model.

    For instance, instances of the following class wouldn't need to mark
    their version or description on initialization:

    .. code-block:: python

       class MyPlugin(Plugin):
           version=0.3
           description='A dull plugin'


    Note that these operations could be included in the __init__ of the
    class, but it would be very inefficient.
    '''
    _subtypes = {}

    def __new__(mcs, name, bases, attrs, **kwargs):
        register_afterwards = False
        defaults = {}
        aliases = {}

        attrs = mcs.expand_with_schema(name, attrs)
        if 'schema' in attrs:
            register_afterwards = True
        for base in bases:
            if hasattr(base, '_defaults'):
                defaults.update(getattr(base, '_defaults'))
            if hasattr(base, '_aliases'):
                aliases.update(getattr(base, '_aliases'))

        info, rest = mcs.split_attrs(attrs)

        for i in list(info.keys()):
            if isinstance(info[i], _Alias):
                aliases[i] = info[i].indict
                if info[i].default is not None:
                    defaults[i] = info[i].default
            else:
                defaults[i] = info[i]

        rest['_defaults'] = defaults
        rest['_aliases'] = aliases

        cls = super(BaseMeta, mcs).__new__(mcs, name, tuple(bases), rest)

        if register_afterwards:
            mcs.register(cls, defaults['@type'])
        return cls

    @classmethod
    def register(mcs, rsubclass, rtype=None):
        mcs._subtypes[rtype or rsubclass.__name__] = rsubclass

    @staticmethod
    def expand_with_schema(name, attrs):
        if 'schema' in attrs:  # Schema specified by name
            schema_file = '{}.json'.format(attrs['schema'])
        elif 'schema_file' in attrs:
            schema_file = attrs['schema_file']
            del attrs['schema_file']
        else:
            return attrs

        if '/' not in 'schema_file':
            thisdir = os.path.dirname(os.path.realpath(__file__))
            schema_file = os.path.join(thisdir,
                                       'schemas',
                                       schema_file)

        schema_path = 'file://' + schema_file

        with open(schema_file) as f:
            schema = json.load(f)

        resolver = jsonschema.RefResolver(schema_path, schema)
        if '@type' not in attrs:
            attrs['@type'] = name
        attrs['_schema_file'] = schema_file
        attrs['schema'] = schema
        attrs['_validator'] = jsonschema.Draft4Validator(schema, resolver=resolver)

        schema_defaults = BaseMeta.get_defaults(attrs['schema'])
        attrs.update(schema_defaults)

        return attrs

    @staticmethod
    def is_func(v):
        return inspect.isroutine(v) or inspect.ismethod(v) or \
            inspect.ismodule(v) or isinstance(v, property)

    @staticmethod
    def is_internal(k):
        return k[0] == '_' or k == 'schema' or k == 'data'

    @staticmethod
    def get_key(key):
        if key[0] != '_':
            key = key.replace("__", ":", 1)
        return key

    @staticmethod
    def split_attrs(attrs):
        '''
        Extract the attributes of the class.

        This allows adding default values in the class definition.
        e.g.:
        '''
        isattr = {}
        rest = {}
        for key, value in attrs.items():
            if not (BaseMeta.is_internal(key)) and (not BaseMeta.is_func(value)):
                isattr[key] = value
            else:
                rest[key] = value
        return isattr, rest

    @staticmethod
    def get_defaults(schema):
        temp = {}
        for obj in [
                schema,
        ] + schema.get('allOf', []):
            for k, v in obj.get('properties', {}).items():
                if 'default' in v and k not in temp:
                    temp[k] = v['default']
        return temp


def make_property(key, default=None):

    def fget(self):
        if default:
            return self.get(key, copy.copy(default))
        return self[key]

    def fdel(self):
        del self[key]

    def fset(self, value):
        self[key] = value

    return fget, fset, fdel


class CustomDict(MutableMapping, object):
    '''
    A dictionary whose elements can also be accessed as attributes. Since some
    characters are not valid in the dot-notation, the attribute names also
    converted. e.g.:

    > d = CustomDict()
    > d.key = d['ns:name'] = 1
    > d.key == d['key']
    True
    > d.ns__name == d['ns:name']
    '''

    _defaults = {}
    _aliases = {'id': '@id'}

    def __init__(self, *args, **kwargs):
        super(CustomDict, self).__init__()
        for k, v in self._defaults.items():
            self[k] = copy.copy(v)
        for arg in args:
            self.update(arg)
        for k, v in kwargs.items():
            self[k] = v
        return self

    def serializable(self, **kwargs):
        def ser_or_down(item):
            if hasattr(item, 'serializable'):
                return item.serializable(**kwargs)
            elif isinstance(item, dict):
                temp = dict()
                for kp in item:
                    vp = item[kp]
                    temp[kp] = ser_or_down(vp)
                return temp
            elif isinstance(item, list) or isinstance(item, set):
                return list(ser_or_down(i) for i in item)
            else:
                return item

        return ser_or_down(self.as_dict(**kwargs))

    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        '''Do not insert data directly, there might be a property in that key. '''
        key = self._key_to_attr(key)
        return setattr(self, key, value)

    def __delitem__(self, key):
        key = self._key_to_attr(key)
        del self.__dict__[key]

    def as_dict(self, verbose=True, aliases=False):
        attrs = self.__dict__.keys()
        if not verbose and hasattr(self, '_terse_keys'):
            attrs = self._terse_keys + ['@type', '@id']
        res = {k: getattr(self, k) for k in attrs
               if not self._internal_key(k) and hasattr(self, k)}
        if not aliases:
            return res
        for k, ok in self._aliases.items():
            if ok in res:
                res[k] = getattr(res, ok)
                del res[ok]
        return res

    def __iter__(self):
        return (k for k in self.__dict__ if not self._internal_key(k))

    def __len__(self):
        return len(self.__dict__)

    def update(self, other):
        for k, v in other.items():
            self[k] = v

    def _attr_to_key(self, key):
        key = key.replace("__", ":", 1)
        key = self._aliases.get(key, key)
        return key

    def _key_to_attr(self, key):
        if self._internal_key(key):
            return key

        if key in self._aliases:
            key = self._aliases[key]
        else:
            key = key.replace(":", "__", 1)
        return key

    def __getattr__(self, key):
        nkey = self._attr_to_key(key)
        if nkey in self.__dict__:
            return self.__dict__[nkey]
        elif nkey == key:
            raise AttributeError("Key not found: {}".format(key))
        return getattr(self, nkey)

    def __setattr__(self, key, value):
        super(CustomDict, self).__setattr__(self._attr_to_key(key), value)

    def __delattr__(self, key):
        super(CustomDict, self).__delattr__(self._attr_to_key(key))

    @staticmethod
    def _internal_key(key):
        return key[0] == '_'

    def __str__(self):
        return json.dumps(self.serializable(), sort_keys=True, indent=4)

    def __repr__(self):
        return json.dumps(self.serializable(), sort_keys=True, indent=4)


_Alias = namedtuple('Alias', ['indict', 'default'])


def alias(key, default=None):
    return _Alias(key, default)



#
#    Copyright 2014 Grupo de Sistemas Inteligentes (GSI) DIT, UPM
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import requests
import logging
from . import models

logger = logging.getLogger(__name__)


class Client(object):
    def __init__(self, endpoint):
        self.endpoint = endpoint

    def analyse(self, input, method='GET', **kwargs):
        return self.request('/', method=method, input=input, **kwargs)

    def evaluate(self, input, method='GET', **kwargs):
        return self.request('/evaluate', method=method, input=input, **kwargs)

    def plugins(self, *args, **kwargs):
        resp = self.request(path='/plugins').plugins
        return {p.name: p for p in resp}

    def datasets(self):
        resp = self.request(path='/datasets').datasets
        return {d.name: d for d in resp}

    def request(self, path=None, method='GET', **params):
        url = '{}{}'.format(self.endpoint.rstrip('/'), path)
        if method == 'POST':
            response = requests.post(url=url, data=params)
        else:
            response = requests.request(method=method, url=url, params=params)

        try:
            resp = models.from_dict(response.json())
        except Exception as ex:
            logger.error(('There seems to be a problem with the response:\n'
                          '\tURL: {url}\n'
                          '\tError: {error}\n'
                          '\t\n'
                          '#### Response:\n'
                          '\tCode: {code}'
                          '\tContent: {content}'
                          '\n').format(
                              error=ex,
                              url=url,
                              code=response.status_code,
                              content=response.content))
            raise ex
        if isinstance(resp, models.Error):
            raise resp
        return resp
